#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import enum
from datetime import datetime
from typing import Tuple, Union, List

import requests
from fim.user import GraphFormat

from fabric_cf.actor.core.common.constants import Constants as CFConstants
from fabric_cf.aits.elements.constants import Constants
from fabric_cf.aits.elements.reservation import ReservationFactory, Reservation
from fabric_cf.aits.elements.slice import SliceFactory, Slice


@enum.unique
class Status(enum.Enum):
    OK = 1
    INVALID_ARGUMENTS = 2
    FAILURE = 3


class OrchestratorHelper:
    HTTP_OK = 200

    def __init__(self):
        self.host = "http://localhost:8700"
        self.token = "eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiIsImtpZCI6ImI0MTUxNjcyMTExOTFlMmUwNWIyMmI1NGIxZDNiNzY2N2U3NjRhNzQ3NzIyMTg1ZTcyMmU1MmUxNDZmZTQzYWEifQ.eyJlbWFpbCI6Imt0aGFyZTEwQGVtYWlsLnVuYy5lZHUiLCJnaXZlbl9uYW1lIjoiS29tYWwiLCJmYW1pbHlfbmFtZSI6IlRoYXJlamEiLCJuYW1lIjoiS29tYWwgVGhhcmVqYSIsImlzcyI6Imh0dHBzOi8vY2lsb2dvbi5vcmciLCJzdWIiOiJodHRwOi8vY2lsb2dvbi5vcmcvc2VydmVyQS91c2Vycy8xMTkwNDEwMSIsImF1ZCI6ImNpbG9nb246L2NsaWVudF9pZC8xMjUzZGVmYzYwYTMyM2ZjYWEzYjQ0OTMyNjQ3NjA5OSIsInRva2VuX2lkIjoiaHR0cHM6Ly9jaWxvZ29uLm9yZy9vYXV0aDIvaWRUb2tlbi82ZmMxYTYyNjY5ZmE0NTk4OTExMjY1ODI0OTgxZThkOC8xNjA2NjU4NjE3NzA4IiwiYXV0aF90aW1lIjoiMTYwNjY1ODYxNyIsImV4cCI6MTYxMDgzMDQzNCwiaWF0IjoxNjEwNzQ0MDM0LCJyb2xlcyI6WyJwcm9qZWN0LWxlYWRzIl0sInByb2plY3RzIjp7IlJFTkNJLVRFU1QiOlsidGFnIDEiLCJ0YWcgMiJdfSwic2NvcGUiOiJhbGwifQ.Gttyz2HQohcj0_fOF00RIGUAMiGtdCBfh5IPs3L2KMKL9Rddy7G_zoqtIynpGr58E8Fn4n4ssGOn9Mutas3kJBCZhF_DawI9uNYGxjhVeinE-Rq7r1Mciwcvj8dA4GPASeurU0yWucioNCxx6u-X4IxbGf2Z01ONPfDN09gSKFk9D_oWy-GTEdiwddr2c_AhtxhCJLS1ZQZ2-mY8R6BJMIZsHU_nv8PJb3luPuRD9b8oVN0W1bIm6JorTja7Tz3P5YcD86ZnBJgPLQW65HoXOfKfxmV4gEHx5c-APjQs1LnbpiL4SvOFdOobgTZYDOV9ei03iXlLdw0GymvQmg53GQ"
        self.headers = {
            'accept': 'application/json',
            'Authorization': f"Bearer {self.token}"
        }
        self.headers_with_content = {
            'accept': 'application/json',
            'Authorization': f"Bearer {self.token}",
            'Content-Type': "text/plain"
        }
        self.ssh_key = "empty ssh string"

    def resources(self, level: int = 1):
        url = f"{self.host}/resources?level={level}"
        return requests.get(url, headers=self.headers, verify=False)

    def portal_resources(self, graph_format: GraphFormat = GraphFormat.JSON_NODELINK):
        url = f"{self.host}/portalresources?graphFormat={graph_format.name}"
        headers = {'accept': 'application/json'}
        return requests.get(url, headers=headers, verify=False)

    def create(self, slice_graph: str, slice_name: str,
               lease_end_time: str = None) -> Tuple[Status, Union[List[Reservation], requests.Response]]:
        if lease_end_time is None:
            url = f"{self.host}/slices/create?sliceName={slice_name}&sshKey={self.ssh_key}"
        else:
            url = f"{self.host}/slices/create?sliceName={slice_name}&sshKey={self.ssh_key}&leaseEndTime={lease_end_time}"
        response = requests.post(url, headers=self.headers_with_content, verify=False, data=slice_graph)
        if response.status_code == self.HTTP_OK:
            reservations = ReservationFactory.create_reservations(reservation_list=
                                                                  response.json()[Constants.PROP_VALUE][Constants.PROP_RESERVATIONS])
            return Status.OK, reservations
        else:
            return Status.FAILURE, response

    def delete(self, slice_id: str) -> Tuple[Status, requests.Response]:
        url = f"{self.host}/slices/delete/{slice_id}"
        response = requests.delete(url, headers=self.headers, verify=False)
        if response.status_code == self.HTTP_OK:
            return Status.OK, response
        else:
            return Status.FAILURE, response

    def slice_status(self, slice_id: str) -> Tuple[Status, Union[Slice, requests.Response]]:
        url = f"{self.host}/slices/status/{slice_id}"
        response = requests.get(url, headers=self.headers, verify=False)
        if response.status_code == self.HTTP_OK:
            slices = SliceFactory.create_slices(slice_list=response.json()[Constants.PROP_VALUE][Constants.PROP_SLICES])
            result = None
            if slices is not None and len(slices) > 0:
                result = next(iter(slices))
            return Status.OK, result
        else:
            return Status.FAILURE, response

    def slices(self, slice_id: str = None,
               state: str = "Active") -> Tuple[Status, Union[List[Slice], str, requests.Response]]:
        url = f"{self.host}/slices?state={state}"
        if slice_id is not None:
            url = f"{self.host}/slices/{slice_id}"

        response = requests.get(url, headers=self.headers, verify=False)
        if response.status_code == self.HTTP_OK:
            if slice_id is None:
                slices = SliceFactory.create_slices(slice_list=response.json()[Constants.PROP_VALUE][Constants.PROP_SLICES])
                return Status.OK, slices
            else:
                slice_graph = response.json()[Constants.PROP_VALUE][Constants.PROP_SLICE_MODEL]
                return Status.OK, slice_graph
        else:
            return Status.FAILURE, response

    def slivers(self, slice_id: str, sliver_id: str = None) -> Tuple[Status, Union[List[Reservation], str, requests.Response]]:
        url = f"{self.host}/slivers?sliceID={slice_id}"
        if sliver_id is not None:
            url = f"{self.host}/slivers/{sliver_id}?sliceID={slice_id}"

        response = requests.get(url, headers=self.headers, verify=False)
        if response.status_code == self.HTTP_OK:
            reservations = ReservationFactory.create_reservations(reservation_list=
                                                                  response.json()[Constants.PROP_VALUE][Constants.PROP_RESERVATIONS])
            return Status.OK, reservations
        else:
            return Status.FAILURE, response

    def sliver_status(self, slice_id: str, sliver_id: str) -> Tuple[Status, Union[Reservation, requests.Response]]:
        url = f"{self.host}/slivers/status/{sliver_id}?sliceID={slice_id}"
        response = requests.get(url, headers=self.headers, verify=False)
        if response.status_code == self.HTTP_OK:
            reservations = ReservationFactory.create_reservations(reservation_list=
                                                                  response.json()[Constants.PROP_VALUE][Constants.PROP_RESERVATIONS])
            result = None
            if reservations is not None and len(reservations) > 0:
                result = next(iter(reservations))
            return Status.OK, result
        else:
            return Status.FAILURE, response

    def renew(self, slice_id: str, new_lease_end_time: str) -> Tuple[Status, Union[requests.Response, None]]:
        """
        Renew a slice
        @param token fabric token
        @param slice_id slice_id
        @param new_lease_end_time new_lease_end_time
        @return Tuple containing Status and List of Reservation Id failed to extend
        """
        response = None
        try:
            # Set the tokens
            url = f"{self.host}/slices/renew/{slice_id}?newLeaseEndTime={new_lease_end_time}"
            response = requests.post(url, headers=self.headers, verify=False)
            if response.status_code != self.HTTP_OK:
                return Status.FAILURE, response

            return Status.OK, response
        except Exception as e:
            return Status.FAILURE, response


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import logging
import time
import unittest
from datetime import datetime

from fim.pluggable import PluggableRegistry
from fim.slivers.attached_components import AttachedComponentsInfo, ComponentSliver, ComponentType
from fim.slivers.capacities_labels import Capacities, CapacityHints
from fim.slivers.instance_catalog import InstanceCatalog
from fim.slivers.network_node import NodeSliver, NodeType

from fabric_cf.actor.core.apis.abc_authority_proxy import ABCAuthorityProxy
from fabric_cf.actor.core.apis.abc_broker_mixin import ABCBrokerMixin
from fabric_cf.actor.core.apis.abc_broker_policy_mixin import ABCBrokerPolicyMixin
from fabric_cf.actor.core.apis.abc_broker_reservation import ABCBrokerReservation
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.apis.abc_slice import ABCSlice
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.core.broker_policy import BrokerPolicy
from fabric_cf.actor.core.delegation.broker_delegation_factory import BrokerDelegationFactory
from fabric_cf.actor.core.kernel.broker_reservation import BrokerReservationFactory
from fabric_cf.actor.core.kernel.resource_set import ResourceSet
from fabric_cf.actor.core.kernel.slice import SliceFactory
from fabric_cf.actor.core.policy.broker_simpler_units_policy import BrokerSimplerUnitsPolicy
from fabric_cf.actor.core.policy.network_node_inventory import NetworkNodeInventory
from fabric_cf.actor.core.proxies.kafka.kafka_authority_proxy import KafkaAuthorityProxy
from fabric_cf.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric_cf.actor.core.time.actor_clock import ActorClock
from fabric_cf.actor.core.time.term import Term
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.resource_type import ResourceType
from fabric_cf.actor.core.util.update_data import UpdateData
from fabric_cf.actor.security.auth_token import AuthToken
from fabric_cf.actor.test.base_test_case import BaseTestCase
from fabric_cf.actor.test.client_callback_helper import ClientCallbackHelper
from fabric_cf.actor.test.fim_test_helper import FimTestHelper
from fabric_cf.actor.test.dummy_authority_proxy import DummyAuthorityProxy


class BrokerSimplerUnitsPolicyTest(BaseTestCase, unittest.TestCase):
    DonateStartCycle = 10
    DonateEndCycle = 100
    logger = logging.getLogger('BrokerSimplerUnitsPolicyTest')
    log_format = '%(asctime)s - %(name)s - {%(filename)s:%(lineno)d} - [%(threadName)s] - %(levelname)s - %(message)s'
    logging.basicConfig(format=log_format, filename="actor.log")
    logger.setLevel(logging.INFO)

    from fabric_cf.actor.core.container.globals import Globals
    Globals.config_file = "./config/config.test.yaml"
    Constants.SUPERBLOCK_LOCATION = './state_recovery.lock'

    from fabric_cf.actor.core.container.globals import GlobalsSingleton
    GlobalsSingleton.get().start(force_fresh=True)
    while not GlobalsSingleton.get().start_completed:
        time.sleep(0.0001)

    adms = None
    broker = None

    def setUp(self) -> None:
        self.adms = FimTestHelper.generate_adms()
        PluggableRegistry.instance.clear()

    def tearDown(self) -> None:
        FimTestHelper.n4j_imp.delete_all_graphs()

    def get_broker(self, *, name: str = BaseTestCase.broker_name, guid: ID = BaseTestCase.broker_guid) -> ABCBrokerMixin:
        db = self.get_container_database()
        db.reset_db()
        self.broker = super().get_broker()
        self.broker.set_recovered(value=True)
        Term.set_clock(self.broker.get_actor_clock())
        return self.broker

    def get_broker_policy(self) -> ABCBrokerPolicyMixin:
        policy = BrokerSimplerUnitsPolicy()
        resource_type = ResourceType(resource_type=NodeType.VM.name)
        inventory = NetworkNodeInventory()
        policy.register_inventory(resource_type=resource_type, inventory=inventory)
        return policy

    def get_request_from_request(self, request: ABCBrokerReservation, units: int, rtype: ResourceType,
                                 start: datetime, end: datetime):
        rset = ResourceSet(units=units, rtype=rtype)

        term = Term(start=request.get_term().get_start_time(), end=end, new_start=start)

        result = BrokerReservationFactory.create(rid=request.get_reservation_id(), resources=rset, term=term,
                                                 slice_obj=request.get_slice())

        result.set_sequence_in(sequence=request.get_sequence_in() + 1)
        return result

    def build_sliver(self) -> NodeSliver:
        node_sliver = NodeSliver()
        node_sliver.resource_type = NodeType.VM
        node_sliver.node_id = "test-slice-node-1"
        cap = Capacities(core=4, ram=64, disk=500)
        catalog = InstanceCatalog()
        instance_type = catalog.map_capacities_to_instance(cap=cap)
        cap_hints = CapacityHints(instance_type=instance_type)
        node_sliver.set_properties(name="node-1", type=NodeType.VM, site="RENC",
                                   capacities=cap, image_type='qcow2', image_ref='default_centos_8',
                                   capacity_hints=cap_hints)
        return node_sliver

    def build_sliver_with_components(self) -> NodeSliver:
        node_sliver = NodeSliver()
        node_sliver.resource_type = NodeType.VM
        node_sliver.node_id = "test-slice-node-1"
        cap = Capacities(core=4, ram=64, disk=500)
        catalog = InstanceCatalog()
        instance_type = catalog.map_capacities_to_instance(cap=cap)
        cap_hints = CapacityHints(instance_type=instance_type)
        node_sliver.set_properties(name="node-1", type=NodeType.VM, site="RENC",
                                   capacities=cap, image_type='qcow2', image_ref='default_centos_8',
                                   capacity_hints=cap_hints)
        component_sliver = ComponentSliver()
        component_sliver.set_properties(type=ComponentType.SmartNIC, model='ConnectX-6', name='nic1')
        node_sliver.attached_components_info = AttachedComponentsInfo()
        node_sliver.attached_components_info.add_device(device_info=component_sliver)

        return node_sliver

    def get_reservation_for_network_node(self, start: datetime, end: datetime, sliver: NodeSliver):
        slice_obj = SliceFactory.create(slice_id=ID(), name="test-slice")
        rtype = ResourceType(resource_type=sliver.resource_type.name)
        rset = ResourceSet(units=1, rtype=rtype, sliver=sliver)
        term = Term(start=start, end=end)
        request = BrokerReservationFactory.create(rid=ID(), resources=rset, term=term, slice_obj=slice_obj)
        return request

    def get_authority_proxy(self) -> ABCAuthorityProxy:
        auth = AuthToken(name="mysite", guid=ID())
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        proxy = KafkaAuthorityProxy(kafka_topic="test-topic", identity=auth, logger=GlobalsSingleton.get().get_logger())

        if ActorRegistrySingleton.get().get_proxy(Constants.PROTOCOL_LOCAL, "mysite") is None:
            dummy = DummyAuthorityProxy(auth=auth)
            ActorRegistrySingleton.get().register_proxy(dummy)
        return proxy

    def check_query_response(self, response: dict, count: int):
        temp = response.get(Constants.QUERY_RESPONSE_STATUS, None)
        self.assertIsNotNone(temp)
        bqm = response.get(Constants.BROKER_QUERY_MODEL, None)
        if count == 0:
            self.assertEqual("", bqm)
        else:
            self.assertIsNotNone(bqm)

    def assert_ticketed(self, r: ABCReservationMixin, units: int, rtype: ResourceType, start: datetime, end: datetime):
        self.assertIsNotNone(r)
        self.assertFalse(r.is_failed())
        self.assertEqual(units, r.get_resources().get_units())
        self.assertEqual(rtype, r.get_resources().get_type())
        self.assertIsNotNone(r.get_term())
        self.assertEqual(start, r.get_term().get_new_start_time())
        self.assertEqual(end, r.get_term().get_end_time())

    def assert_failed(self, r: ABCReservationMixin, update_data: UpdateData = None):
        self.assertIsNotNone(r)
        if update_data is None:
            self.assertTrue(r.is_failed())
        else:
            self.assertTrue(update_data.failed)
        self.assertIsNotNone(r.get_notices())

    def get_source_delegation(self, broker: ABCBrokerMixin, slice_obj: ABCSlice, adm_index: int = 0):
        adm = self.adms[adm_index]
        broker_delegation = BrokerDelegationFactory.create(adm.graph_id, slice_id=slice_obj.get_slice_id(),
                                                           broker=broker)
        broker_delegation.set_graph(graph=adm)
        broker_delegation.set_slice_object(slice_object=slice_obj)
        return broker_delegation

    def test_a_create(self):
        """
        Tests if the actor and the policy can be instantiated.
        """
        self.get_broker()
        self.assertIsNotNone(self.broker)

    def test_b_allocate_ticket(self):
        """
        Requests a ticket for all resources. Checks if the ticket is
        allocated for what was asked. Checks the term. Checks whether the
        reservation is closed when it expires.
        """
        self.broker = self.get_broker()
        controller = self.get_controller()
        clock = self.broker.get_actor_clock()

        proxy = ClientCallbackHelper(name=controller.get_name(), guid=controller.get_guid())
        broker_callback = ClientCallbackHelper(name=self.broker.get_name(), guid=self.broker.get_guid())
        ActorRegistrySingleton.get().register_callback(callback=proxy)
        ActorRegistrySingleton.get().register_callback(callback=broker_callback)
        last_called = proxy.get_called()

        inv_slice = SliceFactory.create(slice_id=ID(), name="inventory-slice")
        inv_slice.set_inventory(value=True)
        source = self.get_source_delegation(self.broker, inv_slice)

        self.broker.register_slice(slice_object=inv_slice)
        self.broker.register_delegation(delegation=source)
        self.broker.donate_delegation(delegation=source)

        cycle = 1
        self.broker.external_tick(cycle=cycle)
        cycle += 1

        units = 1
        start = clock.cycle_start_date(cycle=self.DonateStartCycle)
        end = clock.cycle_end_date(cycle=self.DonateEndCycle - 1)

        sliver = self.build_sliver()
        request = self.get_reservation_for_network_node(start, end, sliver)

        self.broker.ticket(reservation=request, callback=proxy, caller=proxy.get_identity())

        for c in range(cycle, self.DonateEndCycle):
            self.broker.external_tick(cycle=c)
            while self.broker.get_current_cycle() != c:
                time.sleep(0.001)

            if last_called < proxy.get_called():
                self.assert_ticketed(proxy.get_reservation(), units, request.get_type(), start, end)
                last_called = proxy.get_called()

        self.broker.await_no_pending_reservations()

        self.assertEqual(1, proxy.get_called())
        self.assertTrue(request.is_closed())

    def test_c_allocate_ticket2(self):
        """
        Requests a ticket for all resources. Checks if the ticket is
        allocated for what was asked. Checks the term. Checks whether the
        reservation is closed when it expires. Repeat one more time.
        """
        broker = self.get_broker()
        controller = self.get_controller()
        clock = broker.get_actor_clock()

        proxy = ClientCallbackHelper(name=controller.get_name(), guid=controller.get_guid())
        broker_callback = ClientCallbackHelper(name=broker.get_name(), guid=broker.get_guid())
        ActorRegistrySingleton.get().register_callback(callback=proxy)
        ActorRegistrySingleton.get().register_callback(callback=broker_callback)
        last_called = proxy.get_called()

        inv_slice = SliceFactory.create(slice_id=ID(), name="inventory-slice")
        inv_slice.set_inventory(value=True)
        source = self.get_source_delegation(self.broker, inv_slice)

        self.broker.register_slice(slice_object=inv_slice)
        self.broker.register_delegation(delegation=source)
        self.broker.donate_delegation(delegation=source)

        cycle = 1
        broker.external_tick(cycle=cycle)
        cycle += 1

        units = 1
        start = clock.cycle_start_date(cycle=self.DonateStartCycle)
        cycle_end = self.DonateEndCycle - 50
        end = clock.cycle_end_date(cycle=cycle_end)

        sliver = self.build_sliver()
        request = self.get_reservation_for_network_node(start, end, sliver)

        broker.ticket(reservation=request, callback=proxy, caller=proxy.get_identity())

        for c in range(cycle, self.DonateEndCycle):
            broker.external_tick(cycle=c)
            while broker.get_current_cycle() != c:
                time.sleep(0.001)

            if last_called < proxy.get_called():
                self.assert_ticketed(proxy.get_reservation(), units, request.get_type(), start, end)
                last_called = proxy.get_called()

            if c == cycle_end:
                broker.await_no_pending_reservations()
                self.assertTrue(request.is_closed())
                self.assertEqual(1, proxy.get_called())
                start = clock.cycle_start_date(cycle=cycle_end + 1)
                end = clock.cycle_end_date(cycle=self.DonateEndCycle - 1)
                request = self.get_reservation_for_network_node(start, end, sliver=sliver)
                broker.ticket(reservation=request, callback=proxy, caller=proxy.get_identity())

        broker.await_no_pending_reservations()

        self.assertEqual(2, proxy.get_called())
        self.assertTrue(request.is_closed())

    def test_d_extend_ticket(self):
        """
        Requests a ticket for all resources. Checks if the ticket is
        allocated for what was asked. Checks the term. Checks whether the
        reservation is closed when it expires. Repeat one more time.
        """
        broker = self.get_broker()
        controller = self.get_controller()
        clock = broker.get_actor_clock()

        proxy = ClientCallbackHelper(name=controller.get_name(), guid=controller.get_guid())
        broker_callback = ClientCallbackHelper(name=broker.get_name(), guid=broker.get_guid())
        ActorRegistrySingleton.get().register_callback(callback=proxy)
        ActorRegistrySingleton.get().register_callback(callback=broker_callback)
        last_called = proxy.get_called()

        inv_slice = SliceFactory.create(slice_id=ID(), name="inventory-slice")
        inv_slice.set_inventory(value=True)
        source = self.get_source_delegation(self.broker, inv_slice)

        self.broker.register_slice(slice_object=inv_slice)
        self.broker.register_delegation(delegation=source)
        self.broker.donate_delegation(delegation=source)

        cycle = 1
        broker.external_tick(cycle=cycle)
        cycle += 1

        units = 1
        start = clock.cycle_start_date(cycle=self.DonateStartCycle)
        cycle_end = self.DonateEndCycle - 50
        end = clock.cycle_end_date(cycle=cycle_end)

        sliver = self.build_sliver()
        request = self.get_reservation_for_network_node(start, end, sliver)
        reservation = request

        broker.ticket(reservation=request, callback=proxy, caller=proxy.get_identity())

        for c in range(cycle, self.DonateEndCycle):
            broker.external_tick(cycle=c)
            while broker.get_current_cycle() != c:
                time.sleep(0.001)

            if last_called < proxy.get_called():
                self.assert_ticketed(proxy.get_reservation(), units, request.get_type(), start, end)
                last_called = proxy.get_called()

            if c == cycle_end - 10:
                self.assertEqual(1, proxy.get_called())
                start = ActorClock.from_milliseconds(milli_seconds=ActorClock.to_milliseconds(when=end) + 1)
                end = clock.cycle_end_date(cycle=self.DonateEndCycle - 1)
                request = self.get_request_from_request(request, units, request.get_type(), start, end)
                broker.extend_ticket(reservation=request, caller=proxy.get_identity())
                print("Extend done")

        broker.await_no_pending_reservations()

        self.assertEqual(2, proxy.get_called())
        self.assertTrue(reservation.is_closed())

    def test_e_donate(self):
        broker = self.get_broker()
        policy = broker.get_policy()

        slice_obj = SliceFactory.create(slice_id=ID(), name="inventory_slice")
        slice_obj.set_inventory(value=True)

        adm_list_len = len(self.adms)
        adm_list_len -= 1

        self.broker.register_slice(slice_object=slice_obj)
        for i in range(0, adm_list_len):
            source = self.get_source_delegation(self.broker, slice_obj, adm_index=i)
            self.broker.register_delegation(delegation=source)
            self.broker.donate_delegation(delegation=source)

            self.assertEqual(i + 1, len(policy.delegations))

    def test_f_query(self):
        broker = self.get_broker()
        policy = broker.get_policy()

        request = BrokerPolicy.get_broker_query_model_query(level=1)
        response = policy.query(p=request)

        print(response)

        self.check_query_response(response, 0)

        slice_obj = SliceFactory.create(slice_id=ID(), name="inventory_slice")
        slice_obj.set_inventory(value=True)

        adm_list_len = len(self.adms)
        adm_list_len -= 1

        self.broker.register_slice(slice_object=slice_obj)
        for i in range(0, adm_list_len):
            request = BrokerPolicy.get_broker_query_model_query(level=i + 1)
            source = self.get_source_delegation(self.broker, slice_obj, adm_index=i)
            self.broker.register_delegation(delegation=source)
            self.broker.donate_delegation(delegation=source)

            response = policy.query(p=request)
            print(response)
            self.check_query_response(response, i + 1)

    def test_g_advanced_request(self):
        broker = self.get_broker()
        controller = self.get_controller()
        policy = broker.get_policy()
        policy.allocation_horizon = 10

        clock = broker.get_actor_clock()

        proxy = ClientCallbackHelper(name=controller.get_name(), guid=controller.get_guid())
        broker_callback = ClientCallbackHelper(name=broker.get_name(), guid=broker.get_guid())
        ActorRegistrySingleton.get().register_callback(callback=proxy)
        ActorRegistrySingleton.get().register_callback(callback=broker_callback)

        slice_obj = SliceFactory.create(slice_id=ID(), name="inventory_slice")
        slice_obj.set_inventory(value=True)

        source = self.get_source_delegation(self.broker, slice_obj)
        self.broker.register_slice(slice_object=slice_obj)
        self.broker.register_delegation(delegation=source)
        self.broker.donate_delegation(delegation=source)

        cycle = 1
        broker.external_tick(cycle=cycle)
        cycle += 1

        units = 1
        start = clock.cycle_start_date(cycle=self.DonateStartCycle)
        end = clock.cycle_end_date(cycle=self.DonateEndCycle - 1)
        sliver = self.build_sliver_with_components()
        request = self.get_reservation_for_network_node(start, end, sliver=sliver)
        broker.ticket(reservation=request, callback=proxy, caller=proxy.get_identity())

        self.assertEqual(proxy.prepared, 0)
        self.assertEqual(proxy.called, 0)

        for c in range(cycle, self.DonateEndCycle):
            broker.external_tick(cycle=c)
            while broker.get_current_cycle() != c:
                time.sleep(0.001)

            while proxy.prepared != 1:
                time.sleep(0.001)

            self.assert_ticketed(request, units, request.get_type(), start, end)
            self.assertEqual(1, proxy.prepared)

            if proxy.called > 0:
                self.assert_ticketed(proxy.get_reservation(), units, request.get_type(), start, end)

        broker.await_no_pending_reservations()

        self.assertEqual(1, proxy.get_called())
        self.assertTrue(request.is_closed())

    def _test_h_request_insufficient_cores(self):
        broker = self.get_broker()
        controller = self.get_controller()
        policy = broker.get_policy()
        policy.allocation_horizon = 10

        clock = broker.get_actor_clock()

        proxy = ClientCallbackHelper(name=controller.get_name(), guid=controller.get_guid())
        broker_callback = ClientCallbackHelper(name=broker.get_name(), guid=broker.get_guid())
        ActorRegistrySingleton.get().register_callback(callback=proxy)
        ActorRegistrySingleton.get().register_callback(callback=broker_callback)

        slice_obj = SliceFactory.create(slice_id=ID(), name="inventory_slice")
        slice_obj.set_inventory(value=True)

        source = self.get_source_delegation(self.broker, slice_obj)
        self.broker.register_slice(slice_object=slice_obj)
        self.broker.register_delegation(delegation=source)
        self.broker.donate_delegation(delegation=source)

        cycle = 1
        broker.external_tick(cycle=cycle)
        cycle += 1

        start = clock.cycle_start_date(cycle=self.DonateStartCycle)
        end = clock.cycle_end_date(cycle=self.DonateEndCycle - 1)
        sliver = self.build_sliver_with_components()
        caphint = CapacityHints(instance_type="fabric.c64.m384.d4000")
        sliver.set_capacity_hints(caphint=caphint)

        request = self.get_reservation_for_network_node(start, end, sliver=sliver)
        broker.ticket(reservation=request, callback=proxy, caller=proxy.get_identity())

        self.assertEqual(proxy.prepared, 0)
        self.assertEqual(proxy.called, 0)

        for c in range(cycle, self.DonateEndCycle):
            broker.external_tick(cycle=c)
            while broker.get_current_cycle() != c:
                time.sleep(0.001)

            while proxy.prepared != 1:
                time.sleep(0.001)

            self.assert_failed(request)
            self.assertEqual(1, proxy.prepared)

            if proxy.called > 0:
                print("Proxy Called")
            #    self.assert_failed(proxy.get_reservation(), proxy.update_data)

        broker.await_no_pending_reservations()

        self.assertEqual(1, proxy.get_called())
        self.assertTrue(request.is_closed())


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import logging
import traceback
from typing import TYPE_CHECKING, List

import threading

from fabric_mb.message_bus.consumer import AvroConsumerApi
from fabric_mb.message_bus.messages.abc_message_avro import AbcMessageAvro

from fabric_cf.actor.core.common.exceptions import KafkaServiceException

if TYPE_CHECKING:
    from fabric_cf.actor.core.proxies.kafka.services.actor_service import ActorService
    from fabric_cf.actor.core.manage.kafka.services.kafka_actor_service import KafkaActorService


class MessageService(AvroConsumerApi):
    def __init__(self, *, kafka_service: ActorService, kafka_mgmt_service: KafkaActorService, consumer_conf: dict,
                 key_schema_location, value_schema_location: str, topics: List[str], batch_size: int = 5,
                 logger: logging.Logger = None, sync: bool = False):
        super(MessageService, self).__init__(consumer_conf=consumer_conf, key_schema_location=key_schema_location,
                                             value_schema_location=value_schema_location, topics=topics,
                                             batch_size=batch_size, logger=logger, sync=sync)
        self.thread_lock = threading.Lock()
        self.thread = None
        self.kafka_service = kafka_service
        self.kafka_mgmt_service = kafka_mgmt_service

    def start(self):
        try:
            self.thread_lock.acquire()
            if self.thread is not None:
                raise KafkaServiceException("This Message Service has already been started")

            self.thread = threading.Thread(target=self.consume)
            self.thread.setName("MessageService")
            self.thread.setDaemon(True)
            self.thread.start()
            self.logger.debug("Message service has been started")
        finally:
            self.thread_lock.release()

    def stop(self):
        self.shutdown()
        try:
            self.thread_lock.acquire()
            temp = self.thread
            self.thread = None
            if temp is not None:
                self.logger.warning("It seems that the Message Service thread is running. Interrupting it")
                try:
                    temp.join()
                except Exception as e:
                    self.logger.error("Could not join Message Service thread {}".format(e))
                finally:
                    self.thread_lock.release()
        finally:
            if self.thread_lock is not None and self.thread_lock.locked():
                self.thread_lock.release()

    def handle_message(self, message: AbcMessageAvro):
        try:
            if message.get_message_name() == AbcMessageAvro.claim_resources or \
                    message.get_message_name() == AbcMessageAvro.reclaim_resources or \
                    message.get_message_name() == AbcMessageAvro.get_slices_request or \
                    message.get_message_name() == AbcMessageAvro.get_reservations_request or \
                    message.get_message_name() == AbcMessageAvro.get_reservations_state_request or \
                    message.get_message_name() == AbcMessageAvro.get_delegations or \
                    message.get_message_name() == AbcMessageAvro.get_reservation_units_request or \
                    message.get_message_name() == AbcMessageAvro.get_unit_request or \
                    message.get_message_name() == AbcMessageAvro.get_broker_query_model_request or \
                    message.get_message_name() == AbcMessageAvro.add_slice or \
                    message.get_message_name() == AbcMessageAvro.update_slice or \
                    message.get_message_name() == AbcMessageAvro.remove_slice or \
                    message.get_message_name() == AbcMessageAvro.close_reservations or \
                    message.get_message_name() == AbcMessageAvro.update_reservation or \
                    message.get_message_name() == AbcMessageAvro.remove_reservation or \
                    message.get_message_name() == AbcMessageAvro.extend_reservation or \
                    message.get_message_name() == AbcMessageAvro.maintenance_request:
                self.kafka_mgmt_service.process(message=message)
            else:
                self.kafka_service.process(message=message)
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error(e)
            self.logger.error("Discarding the incoming message {}".format(message))



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from fabric_cf.orchestrator.swagger_server.models.base_model_ import Model
from fabric_cf.orchestrator.swagger_server import util


class Status500InternalServerErrorErrors(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    def __init__(self, message: str='Internal Server Error', details: str=None):  # noqa: E501
        """Status500InternalServerErrorErrors - a model defined in Swagger

        :param message: The message of this Status500InternalServerErrorErrors.  # noqa: E501
        :type message: str
        :param details: The details of this Status500InternalServerErrorErrors.  # noqa: E501
        :type details: str
        """
        self.swagger_types = {
            'message': str,
            'details': str
        }

        self.attribute_map = {
            'message': 'message',
            'details': 'details'
        }
        self._message = message
        self._details = details

    @classmethod
    def from_dict(cls, dikt) -> 'Status500InternalServerErrorErrors':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The status_500_internal_server_error_errors of this Status500InternalServerErrorErrors.  # noqa: E501
        :rtype: Status500InternalServerErrorErrors
        """
        return util.deserialize_model(dikt, cls)

    @property
    def message(self) -> str:
        """Gets the message of this Status500InternalServerErrorErrors.


        :return: The message of this Status500InternalServerErrorErrors.
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message: str):
        """Sets the message of this Status500InternalServerErrorErrors.


        :param message: The message of this Status500InternalServerErrorErrors.
        :type message: str
        """

        self._message = message

    @property
    def details(self) -> str:
        """Gets the details of this Status500InternalServerErrorErrors.


        :return: The details of this Status500InternalServerErrorErrors.
        :rtype: str
        """
        return self._details

    @details.setter
    def details(self, details: str):
        """Sets the details of this Status500InternalServerErrorErrors.


        :param details: The details of this Status500InternalServerErrorErrors.
        :type details: str
        """

        self._details = details



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from datetime import datetime
from typing import List

from fim.slivers.base_sliver import BaseSliver
from fim.user.topology import ExperimentTopology

from fabric_cf.actor.core.apis.abc_actor_mixin import ActorType
from fabric_cf.actor.security.fabric_token import FabricToken
from fabric_cf.actor.security.pdp_auth import ActionId, ResourceType, PdpAuth


class AccessChecker:
    """
    Check access for Incoming operation against Policy Decision Point PDP
    """
    @staticmethod
    def check_access(*, action_id: ActionId, token: str,
                     resource: BaseSliver or ExperimentTopology = None,
                     lease_end_time: datetime = None, logger=None) -> FabricToken or None:
        """
        Validates Fabric Token and Check access for Incoming operation against Policy Decision Point PDP
        :param action_id action id
        :param token fabric token
        :param resource resource
        :param lease_end_time lease end time
        :param logger logger

        :returns decoded fabric token on success; throws exception in case of failure
        """
        if token is None:
            return token
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        pdp_config = GlobalsSingleton.get().get_config().get_global_config().get_pdp_config()
        oauth_config = GlobalsSingleton.get().get_config().get_global_config().get_oauth()
        jwt_validator = GlobalsSingleton.get().get_jwt_validator()

        fabric_token = FabricToken(oauth_config=oauth_config, jwt_validator=jwt_validator, logger=logger, token=token)
        fabric_token.validate()

        project, tags = fabric_token.get_project_and_tags()

        AccessChecker.check_pdp_access(action_id=action_id, email=fabric_token.get_email(), project=project,
                                       tags=tags, resource=resource, lease_end_time=lease_end_time, logger=logger)

        return fabric_token

    @staticmethod
    def check_pdp_access(*, action_id: ActionId, email: str, project: str, tags: List[str],
                         resource: BaseSliver or ExperimentTopology = None,
                         lease_end_time: datetime = None, logger=None):
        """
        Check access for Incoming operation against Policy Decision Point PDP
        :param action_id action id
        :param email email
        :param project project
        :param tags tags
        :param resource resource
        :param lease_end_time lease end time
        :param logger logger

        :throws exception in case of failure
        """
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        pdp_config = GlobalsSingleton.get().get_config().get_global_config().get_pdp_config()

        pdp_auth = PdpAuth(config=pdp_config, logger=logger)
        pdp_auth.check_access(email=email, project=project, tags=tags,
                              lease_end_time=lease_end_time,
                              action_id=action_id, resource=resource)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from abc import abstractmethod
from typing import List

from fabric_cf.actor.core.util.client import Client
from fabric_cf.actor.core.util.id import ID


class ClientDatabase:
    """
    This interface describes the methods necessary to maintain information about
    the clients of an actor that acts in a server role.
    """
    @abstractmethod
    def add_client(self, *, client: Client):
        """
        Adds a new database record representing this client
        @param client client
        @throws Exception in case of error
        """

    @abstractmethod
    def update_client(self, *, client: Client):
        """
        Updates database record representing this client
        @param client client
        @throws Exception in case of error
        """

    @abstractmethod
    def remove_client(self, *, guid: ID):
        """
        Removes the specified client record
        @param guid client guid
        @throws Exception in case of error
        """

    @abstractmethod
    def get_client(self, *, guid: ID) -> Client:
        """
        Retrieves the specified client record
        @param guid client guid
        @return vector of properties
        @throws Exception in case of error
        """

    @abstractmethod
    def get_clients(self) -> List[Client]:
        """
        Retrieves all client records
        @return vector of properties
        @throws Exception in case of error
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import pickle
import traceback

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import DatabaseException
from fabric_cf.actor.core.core.unit import Unit
from fabric_cf.actor.core.plugins.db.server_actor_database import ServerActorDatabase
from fabric_cf.actor.core.apis.abc_substrate_database import ABCSubstrateDatabase
from fabric_cf.actor.core.util.id import ID


class SubstrateActorDatabase(ServerActorDatabase, ABCSubstrateDatabase):
    def get_unit(self, *, uid: ID):
        result = None
        try:
            unit_dict = self.db.get_unit(unt_uid=str(uid))
            if unit_dict is not None:
                pickled_unit = unit_dict.get(Constants.PROPERTY_PICKLE_PROPERTIES)
                return pickle.loads(pickled_unit)
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error(e)

        return result

    def add_unit(self, *, u: Unit):
        try:
            if u.get_resource_type() is None:
                raise DatabaseException(Constants.INVALID_ARGUMENT)
            self.lock.acquire()
            if self.get_unit(uid=u.get_id()) is not None:
                self.logger.info("unit {} is already present in database".format(u.get_id()))
                return

            slice_id = str(u.get_slice_id())
            parent = self.get_unit(uid=u.get_parent_id())
            parent_id = None
            if parent is not None:
                parent_id = parent['unt_id']
            res_id = str(u.get_reservation_id())

            properties = pickle.dumps(u)
            self.db.add_unit(slc_guid=slice_id, rsv_resid=res_id,
                             unt_uid=str(u.get_id()), unt_unt_id=parent_id,
                             unt_state=u.get_state().value, properties=properties)
        finally:
            if self.lock.locked():
                self.lock.release()

    def get_units(self, *, rid: ID):
        result = None
        try:
            self.lock.acquire()
            result = []
            unit_dict_list = self.db.get_units(rsv_resid=str(rid))
            if unit_dict_list is not None:
                for u in unit_dict_list:
                    pickled_unit = u.get(Constants.PROPERTY_PICKLE_PROPERTIES)
                    unit_obj = pickle.loads(pickled_unit)
                    result.append(unit_obj)
            return result
        except Exception as e:
            self.logger.error(e)
        finally:
            if self.lock.locked():
                self.lock.release()
        return result

    def remove_unit(self, *, uid: ID):
        try:
            self.lock.acquire()
            self.db.remove_unit(unt_uid=str(uid))
        finally:
            if self.lock.locked():
                self.lock.release()

    def update_unit(self, *, u: Unit):
        try:
            self.lock.acquire()
            properties = pickle.dumps(u)
            self.db.update_unit(unt_uid=str(u.get_id()), properties=properties)
        finally:
            if self.lock.locked():
                self.lock.release()



from fabric_cf.orchestrator.swagger_server.models.slivers import Slivers  # noqa: E501
from fabric_cf.orchestrator.swagger_server.response import slivers_controller as rc


def slivers_get(slice_id):  # noqa: E501
    """Retrieve a listing of user slivers

    Retrieve a listing of user slivers # noqa: E501

    :param slice_id: Slice identifier as UUID
    :type slice_id: str

    :rtype: Slivers
    """
    return rc.slivers_get(slice_id)


def slivers_sliver_id_get(slice_id, sliver_id):  # noqa: E501
    """slivers properties

    Retrieve Sliver properties # noqa: E501

    :param slice_id: Slice identified by universally unique identifier
    :type slice_id: str
    :param sliver_id: Sliver identified by universally unique identifier
    :type sliver_id: str

    :rtype: Slivers
    """
    return rc.slivers_sliver_id_get(slice_id, sliver_id)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import threading
from enum import Enum

from fim.slivers.base_sliver import BaseSliver
from fim.slivers.network_node import NodeSliver

from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.plugins.handlers.config_token import ConfigToken
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.notice import Notice
from fabric_cf.actor.core.util.resource_type import ResourceType


class UnitState(Enum):
    DEFAULT = 0
    PRIMING = 1
    ACTIVE = 2
    MODIFYING = 3
    CLOSING = 4
    CLOSED = 5
    FAILED = 6


class Unit(ConfigToken):
    def __init__(self, *, rid: ID, sliver: BaseSliver = None, slice_id: ID = None, actor_id: ID = None,
                 properties: dict = None, state: UnitState = None, rtype: ResourceType = None):
        # Reservation this unit belongs to (id).
        self.reservation_id = rid
        # Resource type.
        self.rtype = rtype
        # Unique identifier of parent unit (optional).
        self.parent_id = None
        # Properties list.
        self.properties = {}
        if properties is not None:
            self.properties = properties
        # Unit state
        self.state = UnitState.DEFAULT
        if state is not None:
            self.state = state
        # Configuration sequence number. Each unique configuration action is identified by a sequence number.
        self.sequence = 0
        # Slice this unit belongs to.
        self.slice_id = slice_id
        # Actor this unit belongs to.
        self.actor_id = actor_id
        self.notices = Notice()
        # Reservation this unit belongs to.
        self.reservation = None
        # The modified version of the sliver.
        self.modified = None
        self.transfer_out_started = False
        self.sliver = sliver
        self.lock = threading.Lock()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['transfer_out_started']
        del state['reservation']
        del state['lock']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.transfer_out_started = False
        self.reservation = None
        self.lock = threading.Lock()

    def transition(self, *, to_state: UnitState):
        """
        Transition the state
        @param to_state ti state
        """
        self.state = to_state

    def fail(self, *, message: str, exception: Exception = None):
        """
        Fail a unit
        @param message
        @param exception exception
        """
        try:
            self.lock.acquire()
            self.notices.add(msg=message, ex=exception)
            self.transition(to_state=UnitState.FAILED)
        finally:
            self.lock.release()

    def fail_on_modify(self, *, message: str, exception: Exception = None):
        """
        Fail on modify
        @param message message
        @param exception exception
        """
        try:
            self.lock.acquire()
            self.notices.add(msg=message, ex=exception)
            self.transition(to_state=UnitState.ACTIVE)
            self.sliver = self.modified
            self.modified = None
        finally:
            self.lock.release()

    def set_state(self, *, state: UnitState):
        """
        Set state
        @param state state
        """
        try:
            self.lock.acquire()
            self.transition(to_state=state)
        finally:
            self.lock.release()

    def start_close(self):
        """
        Start close on a unit
        """
        try:
            self.lock.acquire()
            self.transition(to_state=UnitState.CLOSING)
            self.transfer_out_started = True
        finally:
            self.lock.release()

    def start_prime(self) -> bool:
        """
        Start priming the unit
        """
        try:
            self.lock.acquire()
            if self.state == UnitState.DEFAULT or self.state == UnitState.PRIMING:
                self.transition(to_state=UnitState.PRIMING)
                return True
        finally:
            self.lock.release()
        return False

    def start_modify(self) -> bool:
        """
        Start modifying
        """
        try:
            self.lock.acquire()
            if self.state == UnitState.ACTIVE or self.state == UnitState.MODIFYING:
                self.transition(to_state=UnitState.MODIFYING)
                return True
        finally:
            self.lock.release()
        return False

    def activate(self):
        """
        Mark the unit as active
        """
        try:
            self.lock.acquire()
            self.state = UnitState.ACTIVE
        finally:
            self.lock.release()

    def close(self):
        """
        Mark the unit as closed
        """
        try:
            self.lock.acquire()
            self.state = UnitState.CLOSED
        finally:
            self.lock.release()

    def get_id(self) -> ID:
        """
        Get unit id
        @return unit id
        """
        return self.reservation_id

    def get_properties(self) -> dict:
        """
        Get unit properties
        @return unit properties
        """
        return self.properties

    def set_property(self, *, name: str, value: str):
        """
        Set property
        @param name name
        @param value value
        """
        try:
            self.lock.acquire()
            self.properties[name] = value
        finally:
            self.lock.release()

    def get_property(self, *, name: str) -> str:
        """
        Get Property
        @param name name
        @return value
        """
        ret_val = None
        try:
            self.lock.acquire()
            if name in self.properties:
                ret_val = self.properties[name]
        finally:
            self.lock.release()
        return ret_val

    def get_state(self) -> UnitState:
        """
        Get Unit state
        @return unit state
        """
        return self.state

    def clone(self):
        """
        Clone a unit
        @return copy of the object
        """
        ret_val = Unit(rid=self.reservation_id, sliver=self.sliver, properties=self.properties, state=self.state)
        return ret_val

    def get_sequence(self) -> int:
        """
        Get Sequence number
        @return sequence number
        """
        return self.sequence

    def increment_sequence(self) -> int:
        """
        Increment sequence number
        """
        try:
            self.lock.acquire()
            self.sequence += 1
        finally:
            self.lock.release()

        ret_val = self.sequence
        return ret_val

    def decrement_sequence(self) -> int:
        """
        Decrement sequence number
        """
        try:
            self.lock.acquire()
            self.sequence -= 1
        finally:
            self.lock.release()
        ret_val = self.sequence
        return ret_val

    def set_reservation(self, *, reservation: ABCReservationMixin):
        """
        Set Reservation
        @param reservation reservation
        """
        try:
            self.lock.acquire()
            self.reservation = reservation
            self.reservation_id = reservation.get_reservation_id()
        finally:
            self.lock.release()

    def set_slice_id(self, *, slice_id: ID):
        """
        Set slice id
        @param slice_id slice id
        """
        try:
            self.lock.acquire()
            self.slice_id = slice_id
        finally:
            self.lock.release()

    def set_actor_id(self, *, actor_id: ID):
        """
        Set actor id
        @param actor_id actor id
        """
        try:
            self.lock.acquire()
            self.actor_id = actor_id
        finally:
            self.lock.release()

    def is_failed(self) -> bool:
        """
        Check fail status
        @return true if unit is failed, false otherwise
        """
        ret_val = self.state == UnitState.FAILED
        return ret_val

    def is_closed(self) -> bool:
        """
        Check close status
        @return true if unit is closed, false otherwise
        """
        ret_val = self.state == UnitState.CLOSED
        return ret_val

    def is_active(self) -> bool:
        """
        Check active status
        @return true if unit is active, false otherwise
        """
        ret_val = self.state == UnitState.ACTIVE
        return ret_val

    def has_pending_action(self) -> bool:
        """
        Check pending action status
        @return true if unit has pending operation, false otherwise
        """
        return self.state == UnitState.MODIFYING or self.state == UnitState.PRIMING or self.state == UnitState.CLOSING

    def set_modified(self, *, modified):
        """
        Set modified unit
        @param modified modified
        """
        try:
            self.lock.acquire()
            self.modified = modified
        finally:
            self.lock.release()

    def get_modified(self):
        """
        Get modified sliver
        @return modified sliver
        """
        return self.modified

    def get_slice_id(self) -> ID:
        """
        Return slice id
        @return slice id
        """
        return self.slice_id

    def get_reservation_id(self) -> ID:
        """
        Get Reservation Id
        @return reservation id
        """
        return self.reservation_id

    def get_reservation(self) -> ABCReservationMixin:
        """
        Get Reservation
        @return reservation
        """
        ret_val = self.reservation
        return ret_val

    def get_parent_id(self) -> ID:
        """
        Get Parent Id
        @return parent guid
        """
        return self.parent_id

    def set_parent_id(self, *, parent_id: ID):
        """
        Set Parent id
        @param parent_id parent id
        """
        try:
            self.lock.acquire()
            self.parent_id = parent_id
        finally:
            self.lock.release()

    def complete_modify(self):
        """
        Complete Modify operation
        """
        try:
            self.lock.acquire()
            self.transition(to_state=UnitState.ACTIVE)
            self.sliver = self.modified
            self.modified = None
        finally:
            self.lock.release()

    def get_actor_id(self) -> ID:
        """
        Get Actor id
        @return actor id
        """
        return self.actor_id

    def get_resource_type(self) -> ResourceType:
        """
        Get Resource Type
        @return resource type
        """
        return self.rtype

    def set_resource_type(self, *, rtype: ResourceType):
        """
        Set resource type
        @param rtype resource type
        """
        self.rtype = rtype

    def __eq__(self, other):
        if not isinstance(other, Unit):
            return NotImplemented

        return self.reservation_id == other.reservation_id

    def get_notices(self) -> Notice:
        """
        Get notices
        @return notices
        """
        ret_val = self.notices
        return ret_val

    def add_notice(self, *, notice: str):
        """
        Add a notice
        @param notice notice
        """
        try:
            self.lock.acquire()
            self.notices.add(msg=notice)
        finally:
            self.lock.release()

    def __str__(self):
        return f"[unit: {self.reservation_id} actor: {self.actor_id} state: {self.state} " \
               f"sliver: {self.sliver} properties: {self.properties}]"

    def __hash__(self):
        return self.reservation_id.__hash__()

    def get_sliver(self) -> BaseSliver:
        """
        Return the sliver associated with the Unit
        :return: sliver
        """
        return self.sliver

    def set_sliver(self, *, sliver: BaseSliver):
        """
        Set the sliver associated with the Unit
        :param sliver: sliver
        :return:
        """
        try:
            self.lock.acquire()
            self.sliver = sliver
        finally:
            self.lock.release()

    def update_sliver(self, *, sliver:BaseSliver):
        """
        Update the sliver associated with the Unit
        Handler Process and AM process have different memory space
        So it requires explicit updates to AM process object with the values updated by Handler Processes
        Side effect of multi-processing in Python

        :param sliver Sliver returned by Handler
        """
        try:
            self.lock.acquire()
            self.sliver.set_label_allocations(sliver.get_label_allocations())
            if isinstance(self.sliver, NodeSliver) and isinstance(sliver, NodeSliver):
                self.sliver.management_ip = sliver.management_ip
                if sliver.attached_components_info is not None:
                    for device in sliver.attached_components_info.devices.values():
                        self.sliver.attached_components_info.devices[device.get_name()].label_allocations = device.label_allocations
        finally:
            self.lock.release()


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric_cf.orchestrator.swagger_server.models import VersionData
from fabric_cf.orchestrator.swagger_server.models.version import Version  # noqa: E501
from fabric_cf.orchestrator.swagger_server import received_counter, success_counter, failure_counter, __API_REFERENCE__
from fabric_cf.orchestrator.swagger_server.response.constants import VERSIONS_PATH, GET_METHOD

from fabric_cf import __VERSION__
from fabric_cf.orchestrator.swagger_server.response.cors_response import cors_500, cors_200


def version_get() -> Version:  # noqa: E501
    """version

    Version # noqa: E501


    :rtype: Version
    """
    received_counter.labels(GET_METHOD, VERSIONS_PATH).inc()
    from fabric_cf.actor.core.container.globals import GlobalsSingleton
    logger = GlobalsSingleton.get().get_logger()
    try:
        version = VersionData()
        version.reference = __API_REFERENCE__
        version.version = __VERSION__
        response = Version()
        response.data = [version]
        response.size = len(response.data)
        response.status = 200
        response.type = 'version'
        success_counter.labels(GET_METHOD, VERSIONS_PATH).inc()
        return cors_200(response_body=response)
    except Exception as exc:
        details = 'Oops! something went wrong with version_get(): {0}'.format(exc)
        logger.error(details)
        failure_counter.labels(GET_METHOD, VERSIONS_PATH).inc()
        return cors_500(details=details)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from abc import abstractmethod
from datetime import datetime

from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
from fabric_cf.actor.core.apis.abc_callback_proxy import ABCCallbackProxy
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.kernel.failed_rpc import FailedRPC
from fabric_cf.actor.core.apis.abc_server_reservation import ABCServerReservation
from fabric_cf.actor.core.apis.abc_slice import ABCSlice
from fabric_cf.actor.core.kernel.reservation import Reservation
from fabric_cf.actor.core.kernel.reservation_states import ReservationPendingStates, ReservationStates
from fabric_cf.actor.core.kernel.resource_set import ResourceSet
from fabric_cf.actor.core.time.term import Term
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.rpc_exception import RPCError
from fabric_cf.actor.core.util.resource_count import ResourceCount
from fabric_cf.actor.core.util.resource_type import ResourceType
from fabric_cf.actor.core.util.update_data import UpdateData
from fabric_cf.actor.security.auth_token import AuthToken


class ReservationServer(Reservation, ABCServerReservation):
    """
    Implementation note on error handling. There are several kinds of errors,
    all of which are logged: - Error on an incoming operation, caught with no
    side effects: these are reflected as exceptions. - Reservation failure.
    Mark reservation as failed, report failure in next update, accept no
    further operations except probe and close, close when ready. - Resource
    failure: log as an event, report notification in next update, release
    failed resources, leave reservation state unchanged, complete operations
    with reduced resource set. Someday obtain new resources to fill the
    deficit. - Probe errors. Log and report. For externally initiated probes,
    always send a lease update. - Asserts. Throw exception if asserts are
    enabled. Errors should generally not throw exceptions up into the calling
    code, with a few exceptions: asserts, and incoming call errors with no
    side effects. Should update exception signatures on all reservation
    classes to reflect this convention.
    """

    def __init__(self, *, rid: ID, resources: ResourceSet, term: Term, slice_object: ABCSlice):
        super().__init__(rid=rid, slice_object=slice_object)
        # Sequence number for incoming messages.
        self.sequence_in = 0
        # Sequence number for outgoing messages.
        self.sequence_out = 0
        # Callback proxy.
        self.callback = None
        # Status of the last server-side operation for the reservation.
        self.update_data = UpdateData()
        # How many update messages have been sent to the client.
        self.update_count = 0
        # Identity of the client actor.
        self.client = None
        # Identity of the server actor.
        self.owner = None
        # Policy in control of the reservation.
        self.policy = None
        self.service_pending = ReservationPendingStates.None_
        self.approved = False
        self.requested_resources = resources
        self.requested_term = term

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['actor']
        del state['logger']
        del state['slice']
        del state['approved']
        del state['previous_resources']
        del state['bid_pending']
        del state['dirty']
        del state['expired']
        del state['pending_recover']
        del state['state_transition']
        del state['service_pending']

        del state['policy']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.actor = None
        self.logger = None
        self.slice = None
        self.approved = False
        self.previous_resources = None
        self.bid_pending = False
        self.dirty = False
        self.expired = False
        self.pending_recover = False
        self.state_transition = False
        self.service_pending = ReservationPendingStates.None_

        self.policy = None

    def restore(self, *, actor: ABCActorMixin, slice_obj: ABCSlice):
        """
        Must be invoked after creating reservation from unpickling
        """
        super().restore(actor=actor, slice_obj=slice_obj)
        if actor is not None:
            if self.callback is not None:
                self.callback.set_logger(logger=actor.get_logger())
            self.policy = actor.get_policy()

    def prepare(self, *, callback: ABCCallbackProxy, logger):
        self.internal_error(err="abstract method")

    def validate_incoming(self):
        if self.slice is None:
            self.error(err="Missing slice")

        if self.requested_resources is None:
            self.error(err="Missing resource set")

        if self.requested_term is None:
            self.error(err="Missing term")

        self.requested_resources.validate_incoming()
        self.requested_term.validate()

    def incoming_request(self):
        """
        Checks reservation state prior to handling an incoming request. These
        checks are not applied to probes or closes.

        @throws Exception
        """
        assert self.slice is not None

        # Disallow a request on a failed reservation, but always send an update to reset the client.
        if self.is_failed():
            self.generate_update()
            self.error(err="server cannot satisfy request (marked failed)")

        # Disallow any further requests on a closing reservation. Generate and update to reset the client.
        if self.is_closed() or self.pending_state == ReservationPendingStates.Closing:
            self.generate_update()
            self.error(err="server cannot satisfy request closing")

    def update_lease(self, *, incoming: ABCReservationMixin, update_data):
        self.internal_error(err="Cannot update a server-side reservation")

    def update_ticket(self, *, incoming: ABCReservationMixin, update_data):
        self.internal_error(err="Cannot update a server-side reservation")

    def handle_failed_rpc(self, *, failed: FailedRPC):
        if failed.get_error_type() == RPCError.NetworkError:
            if self.is_failed() or self.is_closed():
                return

        self.fail(message=f"Failing reservation due to non-recoverable RPC error {failed.get_error_type()}",
                  exception=failed.get_error())

    def clear_notice(self):
        self.update_data.clear()

    def count_when(self, *, when: datetime):
        """
        Counts the number of active and pending units in the reservation at the
        given time instance.

        @param when
                   time instance

        @return counter of active and pending units
        """
        c = Reservation.CountHelper()
        if not self.is_terminal():
            if self.term is not None and self.term.contains(term=when) and self.resources is not None:
                c.active = self.resources.get_concrete_units(when=when)
                c.type = self.resources.type
            else:
                if self.approved and self.approved_term is not None and self.approved_resources is not None:
                    if self.approved_term.contains(term=when):
                        c.active = self.approved_resources.units
                        c.type = self.approved_resources.type
                    else:
                        if self.requested_term is not None and self.requested_term.contains(term=when) and \
                                self.requested_resources is not None:
                            c.pending = self.requested_resources.units
                            c.type = self.requested_resources.type

        return c

    def count(self, *, rc: ResourceCount, when: datetime):
        if self.state == ReservationStates.Nascent or self.state == ReservationStates.Ticketed:
            c = self.count_when(when=when)
            if c.type is not None:
                rc.tally_active(resource_type=c.type, count=c.active)
                rc.tally_pending(resource_type=c.type, count=c.pending)
        elif self.state == ReservationStates.Closed or self.state == ReservationStates.CloseWait:
            if self.resources is not None:
                rc.tally_close(resource_type=self.resources.type, count=self.resources.get_units())
        elif self.state == ReservationStates.Failed and self.resources is not None:
            rc.tally_failed(resource_type=self.resources.type, count=self.resources.get_units())

    def fail(self, *, message: str, exception: Exception = None):
        self.update_data.error(message=message)
        super().fail(message=message, exception=exception)

    def fail_notify(self, *, message: str):
        """
        Fail and notify
        @param message message
        """
        self.fail(message=message)
        self.generate_update()

    def fail_warn(self, *, message: str):
        self.update_data.error(message=message)
        super().fail_warn(message=message)

    @abstractmethod
    def generate_update(self):
        """
        Generates an update to the callback object (if any) for this reservation.
        """

    def get_callback(self) -> ABCCallbackProxy:
        return self.callback

    def get_sequence_in(self) -> int:
        return self.sequence_in

    def get_sequence_out(self) -> int:
        return self.sequence_out

    def get_type(self) -> ResourceType:
        if self.resources is not None:
            return self.resources.type
        else:
            if self.requested_resources is not None:
                return self.requested_resources.type
        return None

    def get_update_data(self) -> UpdateData:
        result = UpdateData()
        result.absorb(other=self.update_data)
        return result

    def get_notices(self) -> str:
        s = super().get_notices()
        notices = self.update_data.get_events()
        if notices is not None:
            s += " {}".format(notices)

        notices = self.update_data.get_message()
        if notices is not None:
            s += " {}".format(notices)

        return s

    def set_owner(self, *, owner: AuthToken):
        self.owner = owner

    def set_requested_resources(self, *, resources: ResourceSet):
        self.requested_resources = resources

    def set_requested_term(self, *, term: Term):
        self.requested_term = term

    def set_sequence_in(self, *, sequence: int):
        self.sequence_in = sequence

    def set_sequence_out(self, *, sequence: int):
        self.sequence_out = sequence

    def get_client_auth_token(self) -> AuthToken:
        if self.callback is not None:
            return self.callback.get_identity()
        else:
            return self.client



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric_mb.message_bus.messages.claim_delegation_avro import ClaimDelegationAvro
from fabric_mb.message_bus.messages.delegation_avro import DelegationAvro
from fabric_mb.message_bus.messages.reclaim_delegation_avro import ReclaimDelegationAvro
from fabric_mb.message_bus.messages.reservation_avro import ReservationAvro
from fabric_mb.message_bus.messages.abc_message_avro import AbcMessageAvro

from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.delegation.delegation_factory import DelegationFactory
from fabric_cf.actor.core.kernel.broker_reservation import BrokerReservationFactory
from fabric_cf.actor.core.kernel.incoming_delegation_rpc import IncomingDelegationRPC
from fabric_cf.actor.core.kernel.incoming_reservation_rpc import IncomingReservationRPC
from fabric_cf.actor.core.kernel.rpc_request_type import RPCRequestType
from fabric_cf.actor.core.proxies.kafka.translate import Translate
from fabric_cf.actor.core.proxies.kafka.services.actor_service import ActorService
from fabric_cf.actor.core.util.id import ID

if TYPE_CHECKING:
    from fabric_mb.message_bus.messages.extend_ticket_avro import ExtendTicketAvro
    from fabric_mb.message_bus.messages.relinquish_avro import RelinquishAvro
    from fabric_mb.message_bus.messages.ticket_avro import TicketAvro
    from fabric_cf.actor.core.apis.abc_broker_reservation import ABCBrokerReservation


class BrokerService(ActorService):
    def pass_agent(self, *, reservation: ReservationAvro) -> ABCBrokerReservation:
        slice_obj = Translate.translate_slice(slice_avro=reservation.slice)
        term = Translate.translate_term_from_avro(term=reservation.term)
        resource_set = Translate.translate_resource_set_from_avro(rset=reservation.resource_set)
        rid = ID(uid=reservation.reservation_id)

        result = BrokerReservationFactory.create(rid=rid, resources=resource_set, term=term, slice_obj=slice_obj,
                                                 actor=self.actor)
        result.set_owner(owner=self.actor.get_identity())
        result.set_sequence_in(sequence=reservation.sequence)

        return result

    def pass_agent_delegation(self, *, delegation: DelegationAvro) -> ABCDelegation:
        slice_obj = Translate.translate_slice(slice_avro=delegation.slice)

        result = DelegationFactory.create(did=delegation.get_delegation_id(), slice_id=slice_obj.get_slice_id())
        result.set_slice_object(slice_object=slice_obj)
        result.set_owner(owner=self.actor.get_identity())
        result.set_sequence_in(value=delegation.sequence)

        return result

    def ticket(self, *, request: TicketAvro):
        rpc = None
        auth_token = Translate.translate_auth_from_avro(auth_avro=request.auth)
        try:
            rsvn = self.pass_agent(reservation=request.reservation)
            callback = self.get_callback(kafka_topic=request.callback_topic, auth=auth_token)
            rpc = IncomingReservationRPC(message_id=ID(uid=request.message_id), request_type=RPCRequestType.Ticket,
                                         reservation=rsvn, callback=callback, caller=auth_token)
        except Exception as e:
            self.logger.error("Invalid Ticket request: {}".format(e))
            raise e
        self.do_dispatch(rpc=rpc)

    def claim_delegation(self, *, request: ClaimDelegationAvro):
        rpc = None
        auth_token = Translate.translate_auth_from_avro(auth_avro=request.auth)
        try:
            dlg = self.pass_agent_delegation(delegation=request.delegation)
            callback = self.get_callback(kafka_topic=request.callback_topic, auth=auth_token)
            rpc = IncomingDelegationRPC(message_id=ID(uid=request.message_id),
                                        request_type=RPCRequestType.ClaimDelegation,
                                        delegation=dlg, callback=callback, caller=auth_token)
        except Exception as e:
            self.logger.error("Invalid Claim request: {}".format(e))
            raise e
        self.do_dispatch(rpc=rpc)

    def reclaim_delegation(self, *, request: ReclaimDelegationAvro):
        rpc = None
        auth_token = Translate.translate_auth_from_avro(auth_avro=request.auth)
        try:
            dlg = self.pass_agent_delegation(delegation=request.delegation)
            callback = self.get_callback(kafka_topic=request.callback_topic, auth=auth_token)
            rpc = IncomingDelegationRPC(message_id=ID(uid=request.message_id),
                                        request_type=RPCRequestType.ReclaimDelegation,
                                        delegation=dlg, callback=callback, caller=auth_token)
        except Exception as e:
            self.logger.error("Invalid reclaim request: {}".format(e))
            raise e
        self.do_dispatch(rpc=rpc)

    def extend_ticket(self, *, request: ExtendTicketAvro):
        rpc = None
        auth_token = Translate.translate_auth_from_avro(auth_avro=request.auth)
        try:
            rsvn = self.pass_agent(reservation=request.reservation)
            callback = self.get_callback(kafka_topic=request.callback_topic, auth=auth_token)
            rpc = IncomingReservationRPC(message_id=ID(uid=request.message_id), request_type=RPCRequestType.ExtendTicket,
                                         reservation=rsvn, caller=auth_token, callback=callback)
        except Exception as e:
            self.logger.error("Invalid extend_ticket request: {}".format(e))
            raise e
        self.do_dispatch(rpc=rpc)

    def relinquish(self, *, request: RelinquishAvro):
        rpc = None
        auth_token = Translate.translate_auth_from_avro(auth_avro=request.auth)
        try:
            rsvn = self.pass_agent(reservation=request.reservation)
            callback = self.get_callback(kafka_topic=request.callback_topic, auth=auth_token)
            rpc = IncomingReservationRPC(message_id=ID(uid=request.message_id), request_type=RPCRequestType.Relinquish,
                                         reservation=rsvn, caller=auth_token, callback=callback)
        except Exception as e:
            self.logger.error("Invalid extend_ticket request: {}".format(e))
            raise e
        self.do_dispatch(rpc=rpc)

    def process(self, *, message: AbcMessageAvro):
        if message.get_message_name() == AbcMessageAvro.ticket:
            self.ticket(request=message)
        elif message.get_message_name() == AbcMessageAvro.claim_delegation:
            self.claim_delegation(request=message)
        elif message.get_message_name() == AbcMessageAvro.reclaim_delegation:
            self.reclaim_delegation(request=message)
        elif message.get_message_name() == AbcMessageAvro.extend_ticket:
            self.extend_ticket(request=message)
        elif message.get_message_name() == AbcMessageAvro.relinquish:
            self.relinquish(request=message)
        else:
            super().process(message=message)



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from fabric_cf.orchestrator.swagger_server.models.base_model_ import Model
from fabric_cf.orchestrator.swagger_server.models.status400_bad_request_errors import Status400BadRequestErrors  # noqa: F401,E501
from fabric_cf.orchestrator.swagger_server import util


class Status400BadRequest(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    def __init__(self, errors: List[Status400BadRequestErrors]=None):  # noqa: E501
        """Status400BadRequest - a model defined in Swagger

        :param errors: The errors of this Status400BadRequest.  # noqa: E501
        :type errors: List[Status400BadRequestErrors]
        """
        self.swagger_types = {
            'errors': List[Status400BadRequestErrors]
        }

        self.attribute_map = {
            'errors': 'errors'
        }
        self._errors = errors

    @classmethod
    def from_dict(cls, dikt) -> 'Status400BadRequest':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The status_400_bad_request of this Status400BadRequest.  # noqa: E501
        :rtype: Status400BadRequest
        """
        return util.deserialize_model(dikt, cls)

    @property
    def errors(self) -> List[Status400BadRequestErrors]:
        """Gets the errors of this Status400BadRequest.


        :return: The errors of this Status400BadRequest.
        :rtype: List[Status400BadRequestErrors]
        """
        return self._errors

    @errors.setter
    def errors(self, errors: List[Status400BadRequestErrors]):
        """Sets the errors of this Status400BadRequest.


        :param errors: The errors of this Status400BadRequest.
        :type errors: List[Status400BadRequestErrors]
        """

        self._errors = errors



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from datetime import datetime

from abc import abstractmethod, ABC
from typing import TYPE_CHECKING

from fim.slivers.base_sliver import BaseSliver

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_authority_proxy import ABCAuthorityProxy
    from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
    from fabric_cf.actor.core.time.term import Term
    from fabric_cf.actor.core.util.notice import Notice
    from fabric_cf.actor.core.util.resource_type import ResourceType
    from fabric_cf.actor.core.apis.abc_base_plugin import ABCBasePlugin


class ABCConcreteSet(ABC):
    """
    IConcreteSet defines the interface for concrete set resources. A
    concrete set is intended to represent a set of resources. For example,
    compute servers, storage servers, network paths, etc.
    Note: each concrete set type should implement a default constructor.

    Concrete sets are single threaded: their methods are invoked by the actor
    main thread and there is no need for internal synchronization. When a concrete set
    method needs to trigger a configuration action, the configuration action should
    be executed on a separate thread. Once the configuration action completes, an event
    should be queued to the actor to be processed on the actor main thread.
    """
    @abstractmethod
    def add(self, *, concrete_set: ABCConcreteSet, configure: bool):
        """
        Adds the passed set to the current set. Optionally triggers configuration
        actions on all added units.
        @param concrete_set : set to add
        @param configure :if true, configuration actions will be triggered for all
                   added units
        @raises Exception in case of error
        """

    @abstractmethod
    def change(self, *, concrete_set: ABCConcreteSet, configure: bool):
        """
        Makes changes to the resources in the concrete set. The incoming concrete
        set represents the state that the current set has to be updated to. The
        implementation must determine what units have been added/removed/modified
        and perform the appropriate actions.
        @param concrete_set : concrete resources representing the new state of the current set
        @param configure : if true, configuration actions will be triggered for all added, removed, or modified units
        @raises Exception thrown if something is wrong
        """

    @abstractmethod
    def close(self):
        """
        Initiates close operations on all resources contained in the set.
        """

    @abstractmethod
    def collect_released(self):
        """
        Collects any released (closed) and/or failed resources.
        @returns a concrete set containing released and or/failed resources
        @raises Exception in case of error
        """

    @abstractmethod
    def get_notices(self) -> Notice:
        """
        Gets a a collection of notices or events pertaining to the underlying
        resources. The event notices are consumed: subsequent calls return only
        new information. May return null.
        """

    @abstractmethod
    def get_site_proxy(self) -> ABCAuthorityProxy:
        """
        Return a proxy or reference for the unique site that owns these
        resources.
        @returns the authority that owns the resources
        @raises Exception in case of error
        """

    @abstractmethod
    def holding(self, *, when: datetime) -> int:
        """
        Returns how many units are in the set at the given time instance.
        @param when: time instance
        @returns how many units will be in the set for at the given time instance.
        """

    @abstractmethod
    def is_active(self) -> bool:
        """
        Checks if the concrete set is active. A concrete set is active if all
        units contained in the set are active.
        @returns true if the set is active
        """

    @abstractmethod
    def modify(self, *, sliver: BaseSliver):
        """
        Updates the units in the current set with information contained in the
        passed sliver.
        @param sliver:  Updated Sliver
        @raises Exception in case of error
        """

    @abstractmethod
    def probe(self):
        """
        Checks the status of pending operations.
        @raises Exception in case of error
        """

    @abstractmethod
    def remove(self, *, concrete_set: ABCConcreteSet, configure: bool):
        """
        Removes the passed set from the current set. Optionally triggers
        configuration actions for all removed units. If the lease term for the
        concrete set has changed, this call must be followed by a call to
        extend(Term).
        @param concrete_set : set to remove
        @param configure : if true, configuration actions will be triggered for all
                      removed units
        @raises Exception in case of error
        """

    @abstractmethod
    def setup(self, *, reservation: ABCReservationMixin):
        """
        Initializes the concrete set with information about the containing
        reservation. This method is called with the manager lock on and hence
        should not block for long periods of time.
        @param reservation : reservation this concrete set belongs to
        """

    @abstractmethod
    def validate_concrete(self, *, rtype: ResourceType, units: int, term: Term):
        """
        Validate that the concrete set matches the abstract resource set
        parameters.
        @param rtype : abstract resources resource type
        @param units : abstract resources units
        @param term : abstract resources term
        @raises Exception in case of error
        """

    @abstractmethod
    def validate_incoming(self):
        """
        Validates a concrete set as it is received by an actor from another
        actor. This method should examine the contents of the concrete set and
        determine whether it is well-formed. Well-formed is an implementation
        specific notion. For example, a ticket is well-formed if all claims
        the ticket is composed of nest properly. This method is also the place to
        perform any additional verification required to ascertain that the
        resources represented by the concrete set are valid. For example, a Sharp
        ticket is valid if it does not result in oversubscription.

        This method is called from ResourceSet with no locks on.

        @raises Exception if validation fails
        """

    @abstractmethod
    def validate_outgoing(self):
        """
        Validates a concrete set as it is about to be sent from the actor to
        another actor. This method should examine the contents of the concrete
        set and determine whether it is well-formed. Any other validation
        required before sending the concrete set should go in this function.

        This method is called from ResourceSet with no locks on.

        This method is called from ResourceSet with no locks on.

        @raises Exception if validation fails
        """

    @abstractmethod
    def restart_actions(self):
        """
        This method will be called during recovery to ensure that all pending
        actions are restarted. If a unit has an outstanding action that has not
        completed yet, that action would have to be restarted during this call.
        @raises Exception in case of error
        """

    @abstractmethod
    def clone_empty(self):
        """
        Makes an "empty" clone of this concrete set. An "empty" clone is a copy
        of a concrete set with the "set" removed from it.
        @returns an "empty" clone of this concrete set
        """

    @abstractmethod
    def clone(self):
        """
        Makes a clone of the concrete set. Unlike clone_empty(), this
        method preserves the set: the set elements are the same objects as the
        original IConcreteSet, but the indexing structures are different. That
        is, adding/removing units to the original should not affect the clone.
        But modifications to an individual unit should be visible form the
        original and the clone.
        @returns a clone of the concrete set
        """

    @abstractmethod
    def get_units(self) -> int:
        """
        Return the number of units
        @return number of units
        """
        return 0

    @abstractmethod
    def restore(self, *, plugin: ABCBasePlugin, reservation: ABCReservationMixin):
        """
        Restore the Concrete set
        @param plugin plugin
        @param reservation reservation
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric_cf.actor.core.apis.abc_mgmt_controller_mixin import ABCMgmtControllerMixin
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates, ReservationPendingStates
from fabric_cf.orchestrator.core.exceptions import OrchestratorException
from fabric_cf.orchestrator.core.reservation_id_with_modify_index import ReservationIDWithModifyIndex
from fabric_cf.orchestrator.core.status_checker import StatusChecker, Status


class ModifyStatusChecker(StatusChecker):
    def __init__(self):
        super().__init__()
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()

    def check_(self, *, controller: ABCMgmtControllerMixin, rid) -> Status:
        """
        Check Status
        :param controller:
        :param rid:
        :return:
        """
        if not isinstance(rid, ReservationIDWithModifyIndex):
            return Status.NOT_READY

        try:
            reservations = controller.get_reservations(rid=rid.get_reservation_id())
            if reservations is None or len(reservations) == 0:
                raise OrchestratorException("Unable to obtain reservation information for {}".format(
                    rid.get_reservation_id()))
            reservation = next(iter(reservations))

            if reservation.get_state() != ReservationStates.Active or \
                    reservation.get_pending_state() != ReservationPendingStates.None_:
                self.logger.debug("Returning NOTREADY for reservation {} in state ({}, {})".format(
                    rid, reservation.get_state(), reservation.get_pending_state()))
                return Status.NOT_READY

            if reservation.get_state() == ReservationStates.Failed or \
                    reservation.get_pending_state() == ReservationStates.Closed:
                return Status.NOT_OK

            units_list = controller.get_reservation_units(rid=rid.get_reservation_id())

            if len(units_list) == 0:
                raise OrchestratorException("No units associated with reservation {}".format(rid.get_reservation_id()))

            unit = units_list.__iter__().__next__()

            properties = unit.get_properties()
            code_property_name = Constants.UNIT_MODIFY_PROP_PREFIX + rid.get_modify_index() + \
                                 Constants.UNIT_MODIFY_PROP_CODE_SUFFIX

            message_property_name = Constants.UNIT_MODIFY_PROP_PREFIX + rid.get_modify_index() + \
                                    Constants.UNIT_MODIFY_PROP_MESSAGE_SUFFIX
            modify_failed = False
            modify_error_message = None
            modify_error_code = 0

            for key, value in properties.items():
                if not modify_failed and code_property_name == key:
                    modify_error_code = int(value)
                    if modify_error_code == 0:
                        self.logger.info("Reservation {} was modified successfully for modify index {}".format(
                            rid.get_reservation_id(), rid.get_modify_index()))
                        return Status.OK
                    else:
                        modify_failed = True
                if message_property_name == key:
                    modify_error_message = value
            if modify_failed:
                self.logger.info("Reservation {} failed modify for modify index {} with code {} and message {}".format(
                    rid.get_reservation_id(), rid.get_modify_index(), modify_error_code, modify_error_message))
                return Status.NOT_OK

        except Exception as e:
            self.logger.error("Exception occurred e: {}".format(e))
        return Status.NOT_READY



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from ctypes import Union
from typing import Tuple

from fim.slivers.capacities_labels import Labels, Capacities
from fim.slivers.delegations import DelegationFormat, Delegations


class InventoryForType:
    def __init__(self):
        self.logger = None

    def set_logger(self, *, logger):
        self.logger = logger

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['logger']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = None

    def _get_delegations(self, *, lab_cap_delegations: Delegations) -> Tuple[str, Union[Labels, Capacities]]:
        # Grab Label Delegations
        delegation_id, deleg = lab_cap_delegations.get_sole_delegation()
        #self.logger.debug(f"Available label/capacity delegations: {deleg} format {deleg.get_format()}")
        # ignore pool definitions and references for now
        if deleg.get_format() != DelegationFormat.SinglePool:
            return None, None
        # get the Labels/Capacities object
        delegated_label_capacity = deleg.get_details()
        return delegation_id, delegated_label_capacity

    @abstractmethod
    def free(self, *, count: int, request: dict = None, resource: dict = None) -> dict:
        """
        Frees the specified number of resource units.
        @param count number of units
        @param request request properties
        @param resource resource properties
        @return new resource properties
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from abc import abstractmethod

from fabric_cf.actor.core.apis.abc_client_reservation import ABCClientReservation
from fabric_cf.actor.core.apis.abc_controller_policy import ABCControllerPolicy
from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import ControllerException
from fabric_cf.actor.core.core.policy import Policy
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates, ReservationPendingStates
from fabric_cf.actor.core.kernel.resource_set import ResourceSet
from fabric_cf.actor.core.time.term import Term
from fabric_cf.actor.core.time.calendar.controller_calendar import ControllerCalendar
from fabric_cf.actor.core.util.reservation_set import ReservationSet


class ControllerCalendarPolicy(Policy, ABCControllerPolicy):
    """
    The base class for all calendar-based service manager policy implementations.
    """
    def __init__(self):
        super().__init__()
        # calendar
        self.calendar = None
        # Contains reservations for which we may have completed performing
        # bookkeeping actions but may need to wait for some other event to take
        # place before we raise the corresponding event.
        self.pending_notify = ReservationSet()
        # If the actor is initialized
        self.initialized = False
        # If true, the orchestrator will close reservations lazily: it will not
        # issue a close and will wait until the site terminates the lease. The
        # major drawback is that leave actions will not be able to connect to the
        # resources, since the resources will not exist at this time.
        self.lazy_close = False

    def check_pending(self):
        """
        Checks pending operations, and installs successfully completed
        requests in the holdings calendar. Note that the policy module must add
        bids to the pending set, or they may not install in the calendar.
       
        @raises Exception in case of error
        """
        rvset = self.calendar.get_pending()
        if rvset is None:
            return

        for reservation in rvset.values():
            if reservation.is_failed():
                # This reservation has failed. Remove it from the list. This is
                # a separate case, because we may fail but not satisfy the
                # condition of the else statement.
                self.logger.debug("Removing failed reservation from the pending list: {}".format(reservation))
                self.calendar.remove_pending(reservation=reservation)
                self.pending_notify.remove(reservation=reservation)
            elif reservation.is_no_pending() and not reservation.is_pending_recover():
                # No pending operation and we are not about the reissue a recovery operation on this reservation.
                self.logger.debug("Controller pending request completed {}".format(reservation))
                if reservation.is_closed():
                    # do nothing handled by close(IReservation)
                    self.logger.debug("No op")
                elif reservation.is_active_ticketed():
                    # An active reservation extended its ticket.
                    # cancel the current close
                    self.calendar.remove_closing(reservation=reservation)
                    # schedule a new close
                    self.calendar.add_closing(reservation=reservation,
                                              cycle=self.get_close(reservation=reservation,
                                                                   term=reservation.get_term()))
                    # Add from start to end instead of close. It is possible
                    # that holdings may not accurately reflect the actual
                    # number of resources towards the end of a lease. This is
                    # because we assume that we still have resources even after
                    # an advanceClose. When looking at this value, see if the
                    # reservation has closed.
                    self.calendar.add_holdings(reservation=reservation,
                                               start=reservation.get_term().get_new_start_time(),
                                               end=reservation.get_term().get_end_time())
                    self.calendar.add_redeeming(reservation=reservation, cycle=self.get_redeem(reservation=reservation))
                    if reservation.is_renewable():
                        cycle = self.get_renew(reservation=reservation)
                        reservation.set_renew_time(time=cycle)
                        reservation.set_dirty()
                        self.calendar.add_renewing(reservation=reservation, cycle=cycle)
                    self.pending_notify.remove(reservation=reservation)
                elif reservation.is_ticketed():
                    # The reservation obtained a ticket for the first time
                    self.calendar.add_holdings(reservation=reservation,
                                               start=reservation.get_term().get_new_start_time(),
                                               end=reservation.get_term().get_end_time())
                    self.calendar.add_redeeming(reservation=reservation,
                                                cycle=self.get_redeem(reservation=reservation))
                    self.calendar.add_closing(reservation=reservation,
                                              cycle=self.get_close(reservation=reservation,
                                                                   term=reservation.get_term()))
                    if reservation.is_renewable():
                        cycle = self.get_renew(reservation=reservation)
                        reservation.set_renew_time(time=cycle)
                        reservation.set_dirty()
                        self.calendar.add_renewing(reservation=reservation, cycle=cycle)
                    self.pending_notify.remove(reservation=reservation)
                elif reservation.get_state() == ReservationStates.Active:
                    if self.pending_notify.contains(reservation=reservation):
                        # We are waiting for transfer in operations to complete
                        # so that we can raise the lease complete event.
                        if reservation.is_active_joined():
                            self.pending_notify.remove(reservation=reservation)
                    else:
                        # Just completed a lease call (redeem or extendLease).
                        # We need to remove this reservation from closing,
                        # because we added it using r.getTerm(), and add this
                        # reservation to closing using r.getLeasedTerm() [the
                        # site could have changed the term of the reservation].
                        # This assumes that r.getTerm has not changed in the
                        # mean time. This is true now, since the state machine
                        # does not allow more than one pending operation.
                        # Should we change this, we will need to update the
                        # code below.
                        self.calendar.remove_closing(reservation=reservation)
                        self.calendar.add_closing(reservation=reservation,
                                                  cycle=self.get_close(reservation=reservation,
                                                                       term=reservation.get_lease_term()))
                        if reservation.get_renew_time() == 0:
                            reservation.set_renew_time(time=(self.actor.get_current_cycle() + 1))
                            reservation.set_dirty()
                            self.calendar.add_renewing(reservation=reservation, cycle=reservation.get_renew_time())

                        if not reservation.is_active_joined():
                            # add to the pending notify list so that we can raise the event
                            # when transfer in operations complete.
                            self.pending_notify.add(reservation=reservation)
                elif reservation.get_state() == ReservationStates.CloseWait or \
                        reservation.get_state() == ReservationStates.Failed:
                    self.pending_notify.remove(reservation=reservation)
                else:
                    self.logger.warning("Invalid state on reservation. We may be still recovering: {}".format(
                        reservation))
                    continue

                if not self.pending_notify.contains(reservation=reservation):
                    self.logger.debug("Removing from pending: {}".format(reservation))
                    self.calendar.remove_pending(reservation=reservation)

    def close(self, *, reservation: ABCReservationMixin):
        # ignore any scheduled/in progress operations
        self.calendar.remove_scheduled_or_in_progress(reservation=reservation)

    def closed(self, *, reservation: ABCReservationMixin):
        # remove the reservation from all calendar structures
        self.calendar.remove_holdings(reservation=reservation)
        self.calendar.remove_redeeming(reservation=reservation)
        self.calendar.remove_renewing(reservation=reservation)
        self.calendar.remove_closing(reservation=reservation)
        self.pending_notify.remove(reservation=reservation)

    def demand(self, *, reservation: ABCClientReservation):
        if not reservation.is_nascent():
            self.logger.error("demand reservation is not fresh")
        else:
            self.calendar.add_demand(reservation=reservation)

    def extend(self, *, reservation: ABCReservationMixin, resources: ResourceSet, term: Term):
        # cancel any previously scheduled extends
        self.calendar.remove_renewing(reservation=reservation)
        # do not cancel the close: the extend may fail cancel any pending redeem: we will redeem after the extension
        self.calendar.remove_redeeming(reservation=reservation)
        # There should be no pending operations for this reservation at this time
        # Add to the pending list so that we can track the progress of the reservation
        self.calendar.add_pending(reservation=reservation)

    def finish(self, *, cycle: int):
        super().finish(cycle=cycle)
        self.calendar.tick(cycle=cycle)

    @abstractmethod
    def get_close(self, *, reservation: ABCClientReservation, term: Term) -> int:
        """
        Returns the time that a reservation should be closed.

        @params reservation reservation
        @params term term

        @returns the close time of the reservation (cycle)

        @raises Exception in case of error
        """

    def get_closing(self, *, cycle: int) -> ReservationSet:
        closing = self.calendar.get_closing(cycle=cycle)
        result = ReservationSet()
        for reservation in closing.values():
            if not reservation.is_failed():
                self.calendar.add_pending(reservation=reservation)
                result.add(reservation=reservation)
            else:
                self.logger.warning("Removing failed reservation from the closing list: {}".format(reservation))
        return result

    @abstractmethod
    def get_redeem(self, *, reservation: ABCClientReservation) -> int:
        """
        Returns the time when the reservation should be redeemed.

        @params reservation the reservation

        @returns the redeem time of the reservation (cycle)

        @raises Exception in case of error
        """

    def get_redeeming(self, *, cycle: int) -> ReservationSet:
        redeeming = self.calendar.get_redeeming(cycle=cycle)
        for reservation in redeeming.values():
            self.calendar.add_pending(reservation=reservation)
        return redeeming

    @abstractmethod
    def get_renew(self, *, reservation: ABCClientReservation) -> int:
        """
        Returns the time when the reservation should be renewed.

        @params reservation the reservation

        @returns the renew time of the reservation (cycle)

        @raises Exception in case of error
        """

    def initialize(self):
        if not self.initialized:
            super().initialize()
            self.calendar = ControllerCalendar(clock=self.clock)
            self.initialized = True

    def is_expired(self, *, reservation: ABCReservationMixin):
        """
        Checks if the reservation has expired.

        @params reservation reservation to check

        @returns true or false
        """
        term = reservation.get_term()
        end = self.clock.cycle(when=term.get_end_time())
        return self.actor.get_current_cycle() > end

    def remove(self, *, reservation: ABCReservationMixin):
        # remove the reservation from the calendar
        self.calendar.remove(reservation=reservation)

    def revisit(self, *, reservation: ABCReservationMixin):
        super().revisit(reservation=reservation)

        if reservation.get_state() == ReservationStates.Nascent:
            self.calendar.add_pending(reservation=reservation)

        elif reservation.get_state() == ReservationStates.Ticketed:

            if reservation.get_pending_state() == ReservationPendingStates.None_:

                if reservation.is_pending_recover():

                    self.calendar.add_pending(reservation=reservation)

                else:

                    self.calendar.add_redeeming(reservation=reservation, cycle=self.get_redeem(reservation=reservation))

                self.calendar.add_holdings(reservation=reservation, start=reservation.get_term().get_new_start_time(),
                                           end=reservation.get_term().get_end_time())

                self.calendar.add_closing(reservation=reservation,
                                          cycle=self.get_close(reservation=reservation, term=reservation.get_term()))

                if reservation.is_renewable() and reservation.get_renew_time() != 0:
                    # Scheduling renewal is a bit tricky, since it may
                    # involve communication with the upstream broker.
                    # However, in some recovery cases, typical in one
                    # container deployment, the broker and the service
                    # manager will be recovering at the same time. In
                    # this case the query may fail and we will have to
                    # fail the reservation.
                    # Our approach here is as follows: we cache the
                    # renew time in the reservation class and persist
                    # it in the database. When we recover, we will
                    # check the renewTime field of the reservation if
                    # it is non-zero, we will use it, otherwise we will
                    # schedule the renew after we get the lease back
                    # from the authority.
                    self.calendar.add_renewing(reservation=reservation, cycle=reservation.get_renew_time())

            elif reservation.get_pending_state() == ReservationPendingStates.Redeeming:
                raise ControllerException(Constants.INVALID_RECOVERY_STATE)

        elif reservation.get_state() == ReservationStates.Active:
            if reservation.get_pending_state() == ReservationPendingStates.None_:
                # pending list
                if reservation.is_pending_recover():
                    self.calendar.add_pending(reservation=reservation)
                # renewing
                if reservation.is_renewable():
                    self.calendar.add_renewing(reservation=reservation, cycle=reservation.get_renew_time())
                # holdings
                self.calendar.add_holdings(reservation=reservation, start=reservation.get_term().get_new_start_time(),
                                           end=reservation.get_term().get_end_time())
                # closing
                self.calendar.add_closing(reservation=reservation,
                                          cycle=self.get_close(reservation=reservation,
                                                               term=reservation.get_lease_term()))

            elif reservation.get_pending_state() == ReservationPendingStates.ExtendingTicket:
                raise ControllerException(Constants.INVALID_RECOVERY_STATE)

        elif reservation.get_state() == ReservationStates.ActiveTicketed:

            if reservation.get_pending_state() == ReservationPendingStates.None_:
                if reservation.is_pending_recover():
                    self.calendar.add_pending(reservation=reservation)
                else:
                    self.calendar.add_redeeming(reservation=reservation, cycle=self.get_redeem(reservation=reservation))

                # holdings
                self.calendar.add_holdings(reservation=reservation, start=reservation.get_term().get_new_start_time(),
                                           end=reservation.get_term().get_end_time())

                # closing
                self.calendar.add_closing(reservation=reservation,
                                          cycle=self.get_close(reservation=reservation,
                                                               term=reservation.get_lease_term()))

                # renewing
                if reservation.is_renewable():
                    self.calendar.add_renewing(reservation=reservation, cycle=reservation.get_renew_time())

            elif reservation.get_pending_state() == ReservationPendingStates.ExtendingLease:
                raise ControllerException(Constants.INVALID_RECOVERY_STATE)

    def ticket_satisfies(self, *, requested_resources: ResourceSet, actual_resources: ResourceSet,
                         requested_term: Term, actual_term: Term):
        return

    def update_ticket_complete(self, *, reservation: ABCClientReservation):
        return

    def update_delegation_complete(self, *, delegation: ABCDelegation, reclaim: bool = False):
        return

    def lease_satisfies(self, *, request_resources: ResourceSet, actual_resources: ResourceSet, requested_term: Term,
                        actual_term: Term):
        return



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import logging

from fabric_mb.message_bus.messages.lease_reservation_avro import LeaseReservationAvro
from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
from fim.user import Labels, ServiceType

from fabric_cf.actor.core.apis.abc_mgmt_controller_mixin import ABCMgmtControllerMixin
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.utils import sliver_to_str
from fabric_cf.orchestrator.core.i_status_update_callback import IStatusUpdateCallback


class ReservationStatusUpdate(IStatusUpdateCallback):
    def __init__(self, *, logger: logging.Logger = None):
        self.logger = logger
        self.l3_ns = [ServiceType.FABNetv6, ServiceType.FABNetv4]

    def success(self, *, controller: ABCMgmtControllerMixin, reservation: ReservationMng):
        """
        Success callback
        :return:
        """
        if not isinstance(reservation, LeaseReservationAvro):
            return
        if reservation.redeem_processors is None:
            return

        ns_sliver = reservation.get_sliver()

        # Only send updates for L3 Network Service Reservations
        if ns_sliver.get_type() not in self.l3_ns:
            return

        self.logger.debug(f"Triggering Updates for Predecessors of res# {reservation.get_reservation_id()}")

        for pred in reservation.get_redeem_predecessors():
            result = controller.get_reservations(rid=ID(uid=pred.get_reservation_id()))
            if result is None or len(result) == 0:
                continue
            pred_res = result[0]

            # Update the sliver
            node_sliver = pred_res.get_sliver()
            modified = False
            for d in node_sliver.attached_components_info.devices.values():
                for ns in d.network_service_info.network_services.values():
                    self.logger.debug(f"Checking keys {ns_sliver.interface_info.interfaces.keys()}")
                    for ifs_name, ifs in ns.interface_info.interfaces.items():
                        if ifs.flags is None or not ifs.flags.auto_config:
                            continue
                        lookup_name = f'{node_sliver.get_name()}-{ifs_name}'
                        if lookup_name in ns_sliver.interface_info.interfaces:
                            ns_ifs = ns_sliver.interface_info.interfaces[lookup_name]
                            if ns_ifs.labels.ipv4 is not None:
                                self.logger.info(f"Updating {ifs} IP4 to {ns_ifs.labels.ipv4}")
                                ifs.label_allocations = Labels.update(ifs.label_allocations,
                                                                           ipv4=ns_ifs.labels.ipv4)
                                modified = True
                            if ns_ifs.labels.ipv6 is not None:
                                self.logger.info(f"Updating {ifs} IP6 to {ns_ifs.labels.ipv6}")
                                ifs.label_allocations = Labels.update(ifs.label_allocations,
                                                                      ipv6=ns_ifs.labels.ipv6)
                                modified = True

            if modified:
                self.logger.debug(f"Updated Node Sliver - {sliver_to_str(sliver=node_sliver)}")
                controller.modify_reservation(rid=ID(uid=pred_res.get_reservation_id()), modified_sliver=node_sliver)

    def failure(self, *, controller: ABCMgmtControllerMixin, reservation: ReservationMng):
        """
        Failure callback
        :return:
        """
        self.logger.debug(f"Reservation {reservation.get_reservation_id()} failed: Nothing to do!")



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_response_handler import ABCResponseHandler
    from fabric_cf.actor.core.kernel.rpc_request import RPCRequest
    from fabric_cf.actor.core.kernel.rpc_request_type import RPCRequestType
    from fabric_cf.actor.core.util.id import ID
    from fabric_cf.actor.core.util.rpc_exception import RPCException, RPCError
    from fabric_cf.actor.security.auth_token import AuthToken


class FailedRPC:
    """
    Represents Failed RPC Message
    """
    def __init__(self, *, e: RPCException = None, request_type: RPCRequestType = None, rid: ID = None,
                 auth: AuthToken = None, request: RPCRequest = None):
        if e is None:
            raise RPCException(message="error cannot be None")

        if rid is None and request is None:
            raise RPCException(message="Both request and rid cannot be None")

        self.error = e
        self.request_type = request_type
        self.rid = rid
        self.remote_auth = auth
        self.request = request

        if request is not None:
            if self.rid is None:
                if request.get_reservation() is not None:
                    self.rid = request.get_reservation().get_reservation_id()
                elif request.get_delegation() is not None:
                    self.rid = request.get_delegation().get_delegation_id()
            if self.request_type is None:
                self.request_type = request.get_request_type()

    def is_reservation_rpc(self) -> bool:
        """
        Check is RPC contains a reservation
        @return true if rid is present, false otherwise
        """
        return self.rid is not None

    def has_request(self) -> bool:
        """
        Check if RPC has request
        @return true if request is present, false otherwise
        """
        return self.request is not None

    def get_request(self) -> RPCRequest:
        """
        Get Request
        @return request
        """
        return self.request

    def get_error(self) -> Exception:
        """
        Get Error
        @return error
        """
        return self.error

    def get_request_type(self) -> RPCRequestType:
        """
        Get Request Type
        @return request type
        """
        return self.request_type

    def get_reservation_id(self) -> ID:
        """
        Get Reservation id
        @return reservation id
        """
        return self.rid

    def get_retry_count(self) -> int:
        """
        Get Retry count
        @return retry count
        """
        return self.request.retry_count

    def get_handler(self) -> ABCResponseHandler:
        """
        Get handler
        @return response handler
        """
        if self.request is None:
            return None
        return self.request.get_handler()

    def get_remote_auth(self) -> AuthToken:
        """
        Get Remote Auth
        @return remote auth token
        """
        return self.remote_auth

    def get_error_type(self) -> RPCError:
        """
        Get error type
        @return error type
        """
        return self.error.get_error_type()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric_cf.orchestrator.core.exceptions import OrchestratorException
from fabric_cf.orchestrator.core.orchestrator_handler import OrchestratorHandler
from fabric_cf.orchestrator.swagger_server import received_counter, success_counter, failure_counter
from fabric_cf.orchestrator.swagger_server.models import Sliver
from fabric_cf.orchestrator.swagger_server.models.slivers import Slivers  # noqa: E501
from fabric_cf.orchestrator.swagger_server.response.constants import GET_METHOD, SLIVERS_GET_PATH, \
    SLIVERS_GET_SLIVER_ID_PATH
from fabric_cf.orchestrator.swagger_server.response.utils import get_token, cors_error_response, cors_success_response


def slivers_get(slice_id) -> Slivers:  # noqa: E501
    """Retrieve a listing of user slivers

    Retrieve a listing of user slivers # noqa: E501

    :param slice_id: Slice identifier as UUID
    :type slice_id: str

    :rtype: Slivers
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels(GET_METHOD, SLIVERS_GET_PATH).inc()
    try:
        token = get_token()
        slivers_dict = handler.get_slivers(slice_id=slice_id, token=token)
        response = Slivers()
        response.data = []
        for s in slivers_dict:
            sliver = Sliver().from_dict(s)
            response.data.append(sliver)
        response.size = len(response.data)
        response.type = "slivers"
        success_counter.labels(GET_METHOD, SLIVERS_GET_PATH).inc()
        return cors_success_response(response_body=response)
    except OrchestratorException as e:
        logger.exception(e)
        failure_counter.labels(GET_METHOD, SLIVERS_GET_PATH).inc()
        return cors_error_response(error=e)
    except Exception as e:
        logger.exception(e)
        failure_counter.labels(GET_METHOD, SLIVERS_GET_PATH).inc()
        return cors_error_response(error=e)


def slivers_sliver_id_get(slice_id, sliver_id) -> Slivers:  # noqa: E501
    """slivers properties

    Retrieve Sliver properties # noqa: E501

    :param slice_id: Slice identified by universally unique identifier
    :type slice_id: str
    :param sliver_id: Sliver identified by universally unique identifier
    :type sliver_id: str

    :rtype: Slivers
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels(GET_METHOD, SLIVERS_GET_SLIVER_ID_PATH).inc()
    try:
        token = get_token()
        slivers_dict = handler.get_slivers(slice_id=slice_id, token=token, sliver_id=sliver_id)
        response = Slivers()
        response.data = []
        for s in slivers_dict:
            sliver = Sliver().from_dict(s)
            response.data.append(sliver)
        response.size = len(response.data)
        response.type = "slivers"
        success_counter.labels(GET_METHOD, SLIVERS_GET_SLIVER_ID_PATH).inc()
        return cors_success_response(response_body=response)
    except OrchestratorException as e:
        logger.exception(e)
        failure_counter.labels(GET_METHOD, SLIVERS_GET_SLIVER_ID_PATH).inc()
        return cors_error_response(error=e)
    except Exception as e:
        logger.exception(e)
        failure_counter.labels(GET_METHOD, SLIVERS_GET_SLIVER_ID_PATH).inc()
        return cors_error_response(error=e)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_callback_proxy import ABCCallbackProxy
    from fabric_cf.actor.core.util.update_data import UpdateData
    from fabric_cf.actor.security.auth_token import AuthToken
    from fabric_cf.actor.core.kernel.resource_set import ResourceSet
    from fabric_cf.actor.core.time.term import Term


class ABCServerReservation(ABCReservationMixin):
    """
    IServerReservation defines the reservation interface for actors acting as servers for other actors.
    """

    @abstractmethod
    def get_callback(self) -> ABCCallbackProxy:
        """
        Returns the callback proxy.
        @return callback proxy
        """

    @abstractmethod
    def get_client_auth_token(self) -> AuthToken:
        """
        Returns the identity of the user represented by the reservation.
        @returns identity of user
        """

    @abstractmethod
    def get_sequence_in(self):
        """
        Returns the sequence number of the last received message.
        @returns sequence number of the last received message
        """

    @abstractmethod
    def get_sequence_out(self):
        """
        Returns the sequence number of the last sent message.
        @returns sequence number of the last sent message
        """

    @abstractmethod
    def is_bid_pending(self) -> bool:
        """
        Checks if the reservation is in progress of being allocated resources
        @returns true if the reservation is in the process of being allocated resources
        """

    @abstractmethod
    def set_bid_pending(self, *, value: bool):
        """
        Indicates whether the reservation is in the progress of
        obtaining resources. This flag should be set (true) when the actor
        policy starts processing a reservation and  should be cleared(false)
        when the policy completes the allocation process.
        @params value : value for the bid pending flag
        """

    @abstractmethod
    def set_sequence_in(self, *, sequence: int):
        """
        Sets the sequence number of the last received message.
        @params sequence : incoming message sequence number
        """

    @abstractmethod
    def set_sequence_out(self, *, sequence: int):
        """
        Sets the sequence number of the last sent message.
        @params sequence : outgoing message sequence number
        """

    @abstractmethod
    def get_update_data(self) -> UpdateData:
        """
        Returns data to be sent back to the client in an update message.
        @return data to be sent back to the client in an update message
        """

    @abstractmethod
    def set_owner(self, *, owner: AuthToken):
        """
        Sets the identity of the server actor that controls the reservation.
        @param owner identity of server actor
        """

    @abstractmethod
    def set_requested_resources(self, *, resources: ResourceSet):
        """
        Sets the requested resources.

        @param resources requested resources
        """

    @abstractmethod
    def set_requested_term(self, *, term: Term):
        """
        Sets the requested term.

        @param term requested term
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import datetime

from abc import abstractmethod, ABC
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fabric_cf.actor.core.time.term import Term
    from fabric_cf.actor.core.util.resource_type import ResourceType
    from fabric_cf.actor.core.kernel.resource_set import ResourceSet
    from fabric_cf.actor.core.util.resource_count import ResourceCount


class ABCReservationResources(ABC):
    """
    IReservationResources defines the API for resources associated with a reservation.
    Each reservation has a number of resource sets associated with it:
        requested: resources that have been requested either to an upstream actor or by a downstream actor.
        approved: resources that have been approved either for making a request to an upstream actor,
                  or to be sent back to a downstream actor to satisfy its request.
        resources: the resources currently bound to the reservation
        previousResources: the previous resource set bound to the reservation
        leasedResources: the concrete resources bound to the reservation.
    """

    @abstractmethod
    def count(self, *, rc: ResourceCount, when: datetime):
        """
        Counts the number of resources in the reservation relative to the specified time.
        The ResourceCount object is updated with the count of active, pending, expired, failed, etc. units.
        Note: "just a hint" unless the kernel lock is held.

        Args:
            rc: holder for counts
            time: when instance
        """

    @abstractmethod
    def get_approved_resources(self) -> ResourceSet:
        """
        Returns the resources approved for this reservation by the last policy decision. If the policy has never
        made a decision about the reservation, this method will return None.

        Returns:
            resources last approved for the reservation. None if no resources have ever been approved.
        """

    @abstractmethod
    def get_approved_term(self) -> Term:
        """
        Returns the term approved for this reservation by the last policy decision. If the policy has never
        made a decision about the reservation, this method will return None.

        Returns:
            term last approved for the reservation. None if no resources have ever been approved.
        """

    @abstractmethod
    def get_approved_type(self) -> ResourceType:
        """
        Returns the resource type approved for this reservation by the last policy decision. If the policy has never
        made a decision about the reservation, this method will return None.

        Returns:
            resource type last approved for the reservation. None if no resources have ever been approved.
        """

    @abstractmethod
    def get_approved_units(self) -> int:
        """
        Returns the number of units approved for this reservation by the last policy decision. If the policy has never
        made a decision about the reservation, this method will return None.

        Returns:
            number of units last approved for the reservation. None if no resources have ever been approved.
        """

    @abstractmethod
    def get_leased_abstract_units(self) -> int:
        """
        Returns number of abstract units leased by the reservation. If the reservation does not represent leased
        resources or has not yet leased any resources, e.g., holds only a ticket, the method will return 0

        Returns:
            number of abstract units leased
        """

    @abstractmethod
    def get_leased_units(self) -> int:
        """
        Returns the number of concrete units leased by the reservation.
        If the reservation does not represent leased resources or has not yet leased any resources,
        e.g., holds only a ticket, the method will return 0.
        Note: This call will always return 0 for reservations that have not recreated their concrete sets, e.g.,
        reservations fetched from the database as a result of a query.
        For such reservations use #get_leased_abstract_units() or obtain the actual reservation object.

        Returns:
            number of leased units
        """

    @abstractmethod
    def get_previous_resources(self) -> ResourceSet:
        """
        Returns the resources represented by/allocated to the reservation at the time before the last update.
        Can be None

        Returns:
            resource represented by the reservation at the time before the last update. Can be None.
        """

    @abstractmethod
    def get_previous_term(self) -> Term:
        """
        Returns the previously allocated term for the reservation.
        Can be None

        Returns:
            previously allocated term. None if reservation has not yet been extended
        """

    @abstractmethod
    def get_requested_resources(self) -> ResourceSet:
        """
        Returns the resources requested for the reservation. If the kernel has not yet issued the resource request
        this method will return None

        Returns:
            resources requested for the reservation. null if no request has been made yet.
        """

    @abstractmethod
    def get_requested_term(self) -> Term:
        """
        Returns the last requested term. If the kernel has not yet issued the resource request
        this method will return None

        Returns:
            last requested term. null if no request has been made yet.
        """

    @abstractmethod
    def get_requested_type(self) -> ResourceType:
        """
        Returns the requested resource type.

        Returns:
            requested resource type
        """

    @abstractmethod
    def get_requested_units(self) -> int:
        """
        Returns the number of requested units. If no units have yet been requested, the method will return 0.

        Returns:
            number of requested units
        """

    @abstractmethod
    def get_resources(self) -> ResourceSet:
        """
        Returns the resources represented by/allocated to the reservation. If no resources have yet been allocated to
        the reservation, this method will return None

        Returns:
            resources represented by the reservation. None if no resources have been allocated to the reservation.
        """

    @abstractmethod
    def get_term(self) -> Term:
        """
        Returns the currently allocated term for the reservation. If no resources have yet been allocated to
        the reservation, this method will return None

        Returns:
            currently allocated term. None if no resources have been allocated to the reservation.
        """

    @abstractmethod
    def get_units(self, *, when: datetime = None) -> int:
        """
        Returns the number of units assigned to the reservation at the specific time instance.
        If the time instance falls outside of the reservation term, this method will return 0.

        Args:
            when: time
        Returns:
            number of units
        """

    @abstractmethod
    def is_approved(self) -> bool:
        """
        Checks if the policy has made a decision for the reservation.

        Returns:
            true if the policy has made a decision for the reservation
        """

    @abstractmethod
    def set_approved(self, *, term: Term = None, approved_resources: ResourceSet = None):
        """
        Sets the term and resources approved for the reservation. This method should be called by the actor policy
        after it determines the resources and term for the reservation. The method also sets the approved flag.

        Args:
            term: term the policy approved
            approved_resources: resources the policy approved
        """

    @abstractmethod
    def set_approved_resources(self, *, approved_resources: ResourceSet):
        """
        Sets the resources approved for the reservation. This method should be called by the actor policy after
        it determines the resources for the reservation. This method will not set the approved flag.
        Once all approval decisions are complete, setApproved()} must be invoked.

        Args:
            approved_resources: resources the policy approved
        """

    @abstractmethod
    def set_approved_term(self, *, term: Term):
        """
        Sets the term approved for the reservation. This method should be called by the actor policy after it
        determines the term for the reservation. This method will not set the approved flag.
        Once all approval decisions are complete, setApproved() must be invoked.

        Args:
            term: approved term
        """
